"""Type resistive.heater.cac.gt, version 000"""

from typing import Literal

from gwproto.named_types import ComponentAttributeClassGt


class ResistiveHeaterCacGt(ComponentAttributeClassGt):
    NameplateMaxPowerW: int
    RatedVoltageV: int
    TypeName: Literal["resistive.heater.cac.gt"] = "resistive.heater.cac.gt"



"""list of all the types"""

from gwproto.named_types.ads111x_based_cac_gt import Ads111xBasedCacGt
from gwproto.named_types.ads111x_based_component_gt import Ads111xBasedComponentGt
from gwproto.named_types.ads_channel_config import AdsChannelConfig
from gwproto.named_types.alert import Alert
from gwproto.named_types.analog_dispatch import AnalogDispatch
from gwproto.named_types.channel_config import ChannelConfig
from gwproto.named_types.channel_readings import ChannelReadings
from gwproto.named_types.component_attribute_class_gt import ComponentAttributeClassGt
from gwproto.named_types.component_gt import ComponentGt
from gwproto.named_types.data_channel_gt import DataChannelGt
from gwproto.named_types.dfr_component_gt import DfrComponentGt
from gwproto.named_types.dfr_config import DfrConfig
from gwproto.named_types.egauge_register_config import EgaugeRegisterConfig
from gwproto.named_types.electric_meter_cac_gt import ElectricMeterCacGt
from gwproto.named_types.electric_meter_channel_config import ElectricMeterChannelConfig
from gwproto.named_types.electric_meter_component_gt import ElectricMeterComponentGt
from gwproto.named_types.fibaro_smart_implant_component_gt import (
    FibaroSmartImplantComponentGt,
)
from gwproto.named_types.fsm_atomic_report import FsmAtomicReport
from gwproto.named_types.fsm_full_report import FsmFullReport
from gwproto.named_types.heartbeat_b import HeartbeatB
from gwproto.named_types.hubitat_component_gt import HubitatComponentGt
from gwproto.named_types.hubitat_poller_component_gt import HubitatPollerComponentGt
from gwproto.named_types.hubitat_tank_component_gt import HubitatTankComponentGt
from gwproto.named_types.i2c_multichannel_dt_relay_component_gt import (
    I2cMultichannelDtRelayComponentGt,
)
from gwproto.named_types.keyparam_change_log import KeyparamChangeLog
from gwproto.named_types.machine_states import MachineStates
from gwproto.named_types.pico_btu_meter_component_gt import PicoBtuMeterComponentGt
from gwproto.named_types.pico_flow_module_component_gt import PicoFlowModuleComponentGt
from gwproto.named_types.pico_tank_module_component_gt import PicoTankModuleComponentGt
from gwproto.named_types.power_watts import PowerWatts
from gwproto.named_types.relay_actor_config import RelayActorConfig
from gwproto.named_types.report import Report
from gwproto.named_types.resistive_heater_cac_gt import ResistiveHeaterCacGt
from gwproto.named_types.resistive_heater_component_gt import ResistiveHeaterComponentGt
from gwproto.named_types.rest_poller_component_gt import RESTPollerComponentGt
from gwproto.named_types.send_snap import SendSnap
from gwproto.named_types.single_reading import SingleReading
from gwproto.named_types.spaceheat_node_gt import SpaceheatNodeGt
from gwproto.named_types.synced_readings import SyncedReadings
from gwproto.named_types.synth_channel_gt import SynthChannelGt
from gwproto.named_types.tank_module_params import TankModuleParams
from gwproto.named_types.ticklist_hall import TicklistHall
from gwproto.named_types.ticklist_hall_report import TicklistHallReport
from gwproto.named_types.ticklist_reed import TicklistReed
from gwproto.named_types.ticklist_reed_report import TicklistReedReport
from gwproto.named_types.web_server_component_gt import WebServerComponentGt

__all__ = [
    "Ads111xBasedCacGt",
    "Ads111xBasedComponentGt",
    "AdsChannelConfig",
    "Alert",
    "AnalogDispatch",
    "ChannelConfig",
    "ChannelReadings",
    "ComponentAttributeClassGt",
    "ComponentGt",
    "DataChannelGt",
    "DfrComponentGt",
    "DfrConfig",
    "EgaugeRegisterConfig",
    "ElectricMeterCacGt",
    "ElectricMeterChannelConfig",
    "ElectricMeterComponentGt",
    "FibaroSmartImplantComponentGt",
    "FsmAtomicReport",
    "FsmFullReport",
    "HeartbeatB",
    "HubitatComponentGt",
    "HubitatPollerComponentGt",
    "HubitatTankComponentGt",
    "I2cMultichannelDtRelayComponentGt",
    "KeyparamChangeLog",
    "MachineStates",
    "PicoBtuMeterComponentGt",
    "PicoFlowModuleComponentGt",
    "PicoTankModuleComponentGt",
    "PowerWatts",
    "RESTPollerComponentGt",
    "RelayActorConfig",
    "Report",
    "ResistiveHeaterCacGt",
    "ResistiveHeaterComponentGt",
    "SendSnap",
    "SingleReading",
    "SpaceheatNodeGt",
    "SyncedReadings",
    "SynthChannelGt",
    "TankModuleParams",
    "TicklistHall",
    "TicklistHallReport",
    "TicklistReed",
    "TicklistReedReport",
    "WebServerComponentGt",
    "cacs",  # noqa: F822
    "components",  # noqa: F822
]



from collections.abc import Mapping
from typing import Any, Optional

import yarl

from gwproto.data_classes.components import HubitatComponent
from gwproto.data_classes.components.component import Component
from gwproto.data_classes.resolver import ComponentResolver
from gwproto.data_classes.sh_node import ShNode
from gwproto.named_types.component_attribute_class_gt import ComponentAttributeClassGt
from gwproto.named_types.hubitat_component_gt import (
    HubitatComponentGt,
    HubitatRESTResolutionSettings,
)
from gwproto.named_types.hubitat_tank_component_gt import HubitatTankComponentGt
from gwproto.named_types.hubitat_tank_gt import (
    FibaroTempSensorSettings,
    FibaroTempSensorSettingsGt,
)


class HubitatTankComponent(
    Component[HubitatTankComponentGt, ComponentAttributeClassGt], ComponentResolver
):
    hubitat: HubitatComponentGt
    devices_gt: list[FibaroTempSensorSettingsGt]
    devices: list[FibaroTempSensorSettings]

    def __init__(
        self, gt: HubitatTankComponentGt, cac: ComponentAttributeClassGt
    ) -> None:
        super().__init__(gt, cac)
        # Create self.hubitat as a proxy containing only the id
        # of the hubitat; the actual component data will be resolved
        # when resolve() is called; Here in the constructor we cannot
        # rely on the actual HubitatComponentGt existing yet.
        self.hubitat = HubitatComponentGt.make_stub(self.gt.Tank.hubitat_component_id)
        self.devices_gt = list(self.gt.Tank.devices)
        self.devices = []

    @property
    def sensor_supply_voltage(self) -> float:
        return self.gt.Tank.sensor_supply_voltage

    @property
    def default_poll_period_seconds(self) -> Optional[float]:
        return self.gt.Tank.default_poll_period_seconds

    @property
    def web_listen_enabled(self) -> bool:
        return self.gt.Tank.web_listen_enabled

    def resolve(
        self,
        tank_node_name: str,
        nodes: dict[str, ShNode],
        components: Mapping[str, Component[Any, Any]],
    ) -> None:
        hubitat_component = components.get(self.hubitat.ComponentId, None)
        if not isinstance(hubitat_component, HubitatComponent):
            raise ValueError(  # noqa: TRY004
                f"ERROR. Component for {self.hubitat.ComponentId} "
                f"has type <{type(hubitat_component)}>. Expected <HubitatComponent>"
            )
        hubitat_settings = HubitatRESTResolutionSettings(hubitat_component.gt)
        devices = [
            FibaroTempSensorSettings.create(
                tank_name=tank_node_name,
                settings_gt=device_gt,
                hubitat=hubitat_settings,
                default_poll_period_seconds=self.default_poll_period_seconds,
            )
            for device_gt in self.devices_gt
            if device_gt.enabled
        ]
        for device in devices:
            if device.node_name not in nodes:
                raise ValueError(
                    f"ERROR. Node not found for tank temp sensor <{device.node_name}>"
                )
        # replace proxy hubitat component, which only had component id.
        # with the actual hubitat component containing data.
        self.hubitat = hubitat_component.gt
        self.devices = devices

        # register voltage attribute for fibaros which accept web posts
        if self.web_listen_enabled and hubitat_component.gt.Hubitat.WebListenEnabled:
            for device in self.devices:
                if device.web_listen_enabled:
                    hubitat_component.add_web_listener(tank_node_name)

    def urls(self) -> dict[str, Optional[yarl.URL]]:
        urls = self.hubitat.urls()
        for device in self.devices:
            urls[device.node_name] = device.url
        return urls



from enum import auto

from gw.enums import GwStrEnum


class FsmActionType(GwStrEnum):
    """
    A list of the finite state machine Actions that a spaceheat node might take. An Action,
    in this context, is a side-effect of a state machine transition that impacts the real world
    (i.e., a relay is actuated).
    Values:
      - RelayPinSet
      - Analog010VSignalSet
      - Analog420maSignalSet

    For more information:
      - [ASLs](https://gridworks-type-registry.readthedocs.io/en/latest/)
      - [Global Authority](https://gridworks-type-registry.readthedocs.io/en/latest/enums.html#shfsmactiontype)
      - [More Info](https://gridworks-protocol.readthedocs.io/en/latest/finite-state-machines.html)
    """

    RelayPinSet = auto()
    Analog010VSignalSet = auto()
    Analog420maSignalSet = auto()

    @classmethod
    def default(cls) -> "FsmActionType":
        return cls.RelayPinSet

    @classmethod
    def values(cls) -> list[str]:
        return [elt.value for elt in cls]

    @classmethod
    def enum_name(cls) -> str:
        return "sh.fsm.action.type"

    @classmethod
    def enum_version(cls) -> str:
        return "000"



from pydantic import ConfigDict

from gwproto.data_classes.sh_node import ShNode
from gwproto.named_types import SynthChannelGt


class SynthChannel(SynthChannelGt):
    created_by_node: ShNode

    model_config = ConfigDict(arbitrary_types_allowed=True, use_enum_values=True)

    def __hash__(self) -> int:
        return hash(self.Id)

    def to_gt(self) -> SynthChannelGt:
        # Copy the current instance excluding the extra fields
        return SynthChannelGt(**self.model_dump(exclude={"created_by_node"}))



"""SCADA Component Class Definition"""

from typing import Generic, TypeVar

from gwproto.named_types import ComponentAttributeClassGt, ComponentGt

ComponentT = TypeVar("ComponentT", bound=ComponentGt)
CacT = TypeVar("CacT", bound=ComponentAttributeClassGt)


class Component(Generic[ComponentT, CacT]):
    gt: ComponentT
    cac: CacT

    def __init__(self, gt: ComponentT, cac: CacT) -> None:
        self.gt = gt
        self.cac = cac

    def __repr__(self) -> str:
        return f"<{self.gt.DisplayName}>  ({self.cac.MakeModel})"


class ComponentOnly(Component[ComponentGt, ComponentAttributeClassGt]): ...



"""Type electric.meter.component.gt, version 001"""

from collections.abc import Sequence
from typing import Literal, Optional, Self

from pydantic import PositiveInt, field_validator, model_validator

from gwproto.named_types import ComponentGt
from gwproto.named_types.electric_meter_channel_config import ElectricMeterChannelConfig


class ElectricMeterComponentGt(ComponentGt):
    ModbusHost: Optional[str] = None
    ModbusPort: Optional[PositiveInt] = None
    ConfigList: Sequence[ElectricMeterChannelConfig]
    TypeName: Literal["electric.meter.component.gt"] = "electric.meter.component.gt"
    Version: str = "001"

    @field_validator("ConfigList")
    @classmethod
    def check_electric_meter_config_list(
        cls, v: list[ElectricMeterChannelConfig]
    ) -> list[ElectricMeterChannelConfig]:
        """
        Axiom 1: Channel Name uniqueness. Data Channel names are
        unique in the config list

        Axiom 2: Egauge Config consistency. If one of the ElectricMeterChannelConfigs
        has an EgaugeRegisterConfig, then they all do.
        """
        # Implement Axiom(s)
        return v

    @model_validator(mode="after")
    def check_axiom_1(self) -> Self:
        """
        Axiom 3: Modbus consistency.
        ModbusHost is None if and only if ModbusPort is None
        """
        # Implement check for axiom 1"
        return self

    @model_validator(mode="after")
    def check_axiom_2(self) -> Self:
        """
        Axiom 4: Egauge4030 Means Modbus.
        If any of the ElectricMeterChannelConfigs have EgaugeRegisterConfig, then the ModbusHost
        is not None
        """
        # Implement check for axiom 2"
        return self



"""Type pico.tank.reader.component.gt, version 010"""

from typing import Literal, Optional

from pydantic import ConfigDict, model_validator
from typing_extensions import Self

from gwproto.enums import TempCalcMethod
from gwproto.named_types.component_gt import ComponentGt


class PicoTankModuleComponentGt(ComponentGt):
    Enabled: bool
    PicoHwUid: Optional[str] = None
    PicoAHwUid: Optional[str] = None
    PicoBHwUid: Optional[str] = None
    TempCalcMethod: TempCalcMethod
    ThermistorBeta: int
    SendMicroVolts: bool
    Samples: int
    NumSampleAverages: int
    PicoKOhms: Optional[int] = None
    SerialNumber: str = "NA"
    AsyncCaptureDeltaMicroVolts: int
    TypeName: Literal["pico.tank.module.component.gt"] = "pico.tank.module.component.gt"
    Version: str = "010"

    model_config = ConfigDict(extra="allow")

    @model_validator(mode="after")
    def check_axiom_1(self) -> Self:
        """
        Axiom 1: PicoHwUid exists  XOR (both PicoAHwUid and PicoBHwUid exist)
        """
        if self.PicoHwUid is not None:
            if self.PicoAHwUid or self.PicoBHwUid:
                raise ValueError(
                    "Can't have both PicoHwUid and any of (PicoAHwUid, PicoBHwUid"
                )
        elif not (self.PicoAHwUid and self.PicoBHwUid):
            raise ValueError(
                "If PicoHwUid is not set, PicoAHwUid and PicoBHwUid must both be set!"
            )

        return self

    @model_validator(mode="after")
    def check_axiom_2(self) -> Self:
        """
        Axiom 2: PicoKOhms exists iff TempCalcMethod is TempCalcMethod.SimpleBetaForPico
        # note this is a known incorrect method, but there are a few in the field
        # that do this.
        """
        is_simple_beta = self.TempCalcMethod == TempCalcMethod.SimpleBetaForPico
        has_kohms = self.PicoKOhms is not None

        if is_simple_beta != has_kohms:
            raise ValueError(
                "PicoKOhms must be provided if and only if TempCalcMethod is SimpleBetaForPico"
            )

        return self



"""Type component.attribute.class.gt, version 001"""

from typing import Optional

from pydantic import BaseModel, ConfigDict, PositiveInt, model_validator
from typing_extensions import Self

from gwproto.enums import MakeModel
from gwproto.property_format import UUID4Str
from gwproto.type_helpers import CACS_BY_MAKE_MODEL


class ComponentAttributeClassGt(BaseModel):
    ComponentAttributeClassId: UUID4Str
    DisplayName: Optional[str] = None
    MakeModel: MakeModel
    MinPollPeriodMs: Optional[PositiveInt] = None
    TypeName: str = "component.attribute.class.gt"
    Version: str = "001"

    model_config = ConfigDict(use_enum_values=True, extra="allow")

    @model_validator(mode="after")
    def check_axiom_1(self) -> Self:
        """
        Axiom 1: Component Attribute Classes captured by MakeModel
                If cac is a ComponentAttributeClassGt,
        then
           - EITHER  its (MakeModel, ComponentAttributeClassId) must be a key,value pair in
        CACS_BY_MAKE_MODEL (below)
           - XOR its MakeModel is MakeModel.UNKNOWNMAKE__UNKNOWNMODEL

        CACS_BY_MAKE_MODEL: Dict[MakeModel, str] = {
            MakeModel.EGAUGE__4030: '739a6e32-bb9c-43bc-a28d-fb61be665522',
            MakeModel.NCD__PR814SPST: 'c6e736d8-8078-44f5-98bb-d72ca91dc773',
            MakeModel.ADAFRUIT__642: '43564cd2-0e78-41a2-8b67-ad80c02161e8',
            MakeModel.GRIDWORKS__WATERTEMPHIGHPRECISION: '7937eb7e-24d5-4d52-990f-cca063484df9',
            MakeModel.GRIDWORKS__SIMPM1: '28897ac1-ea42-4633-96d3-196f63f5a951',
            MakeModel.SCHNEIDERELECTRIC__IEM3455: '6bcdc388-de10-40e6-979a-8d66bfcfe9ba',
            MakeModel.GRIDWORKS__SIMBOOL30AMPRELAY: '69f101fc-22e4-4caa-8103-50b8aeb66028',
            MakeModel.OPENENERGY__EMONPI: '357b9b4f-2550-4380-aa6b-d2cd9c7ba0f9',
            MakeModel.GRIDWORKS__SIMTSNAP1: 'b9f7135e-07a9-42f8-b847-a9bb3ea3770a',
            MakeModel.ATLAS__EZFLO: '13d916dc-8764-4b16-b85d-b8ead3e2fc80',
            MakeModel.HUBITAT__C7__LAN1: '62528da5-b510-4ac2-82c1-3782842eae07',
            MakeModel.GRIDWORKS__TANK_MODULE_1: '60ac199d-679a-49f7-9142-8ca3e6428a5f',
            MakeModel.FIBARO__ANALOG_TEMP_SENSOR: '7ce0ce69-14c6-4cb7-a33f-2aeca91e0680',
            MakeModel.AMPHENOL__NTC_10K_THERMISTOR_MA100GG103BN: '2821c81d-054d-4003-9b07-2c295aef40f5',
            MakeModel.YHDC__SCT013100: '812761ba-6544-4796-9aad-e1c979f58734',
            MakeModel.MAGNELAB__SCT0300050: 'cf312bd6-7ca5-403b-a61b-b2e817ea1e22',
            MakeModel.GRIDWORKS__MULTITEMP1: '432073b8-4d2b-4e36-9229-73893f33f846',
            MakeModel.KRIDA__EMR16I2CV3: '018d9ffb-89d1-4cc4-95c0-f170711b5ffa',
            MakeModel.OMEGA__FTB8007HWPT: '8cf6c726-e38a-4900-9cfe-ae6f053aafdf',
            MakeModel.ISTEC_4440: '62ed724c-ba62-4302-ae30-d52b20d42ad9',
            MakeModel.OMEGA__FTB8010HWPT: 'd9f225f8-eeb5-4cb7-b314-5551b925ea27',
            MakeModel.BELIMO__BALLVALVE232VS: 'a2236d8c-7c9b-403f-9c55-733c62971d09',
            MakeModel.BELIMO__DIVERTERB332L: 'f3261ed0-3fb1-4def-b60b-246960bf85ef',
            MakeModel.TACO__0034EPLUS: '3880ba73-61e5-4b35-9df1-e154a03a3335',
            MakeModel.TACO__007E: '198ebac8-e0b9-4cee-ae91-2ee6db708491',
            MakeModel.ARMSTRONG__COMPASSH: 'ff6863e1-d5f7-4066-8579-2768162321a6',
            MakeModel.HONEYWELL__T6ZWAVETHERMOSTAT: '03533a1f-3cb9-4a1f-8d57-690c0ad0475b',
            MakeModel.PRMFILTRATION__WM075: '61d5c12d-eeca-4835-9a11-e61167d82e0d',
            MakeModel.BELLGOSSETT__ECOCIRC20_18: '0d2ccc36-d2b8-405d-a257-3917111607c5',
            MakeModel.TEWA__TT0P10KC3T1051500: '20779dbb-0302-4c36-9d60-e1962857c2f3',
            MakeModel.EKM__HOTSPWM075HD: 'e52cb571-913a-4614-90f4-5cc81f8e7fe5',
            MakeModel.GRIDWORKS__SIMMULTITEMP: '627ac482-24fe-46b2-ba8c-3d6f1e1ee069',
            MakeModel.GRIDWORKS__SIMTOTALIZER: 'a88f8f4c-fe1e-4645-a7f4-249912131dc8',
            MakeModel.KRIDA__DOUBLEEMR16I2CV3: '29eab8b1-100f-4230-bb44-3a2fcba33cc3',
            MakeModel.GRIDWORKS__TANKMODULE3: 'cbe49338-5d14-47ff-b05e-22031876962e',
        }
        """
        if (
            self.MakeModel not in CACS_BY_MAKE_MODEL
            and self.MakeModel is not MakeModel.default().value
        ):
            raise ValueError(
                "Axiom 1 violated! If MakeModel not in this list, "
                f"must be UNKNOWN: {CACS_BY_MAKE_MODEL}"
            )
        if self.MakeModel is MakeModel.default().value:
            if self.ComponentAttributeClassId in CACS_BY_MAKE_MODEL.values():
                raise ValueError(
                    f"Id {self.ComponentAttributeClassId} already used by known MakeModel!"
                )
        elif self.ComponentAttributeClassId != CACS_BY_MAKE_MODEL[self.MakeModel]:
            raise ValueError(
                f"Axiom 1 violated! MakeModel {self.MakeModel} must have "
                f"id {CACS_BY_MAKE_MODEL[self.MakeModel]}!"
            )
        return self



from enum import auto

from gw.enums import GwStrEnum


class Unit(GwStrEnum):
    """
    Specifies the physical unit of sensed data reported by SCADA
    Values:
      - Unknown
      - Unitless
      - W
      - Celcius
      - Fahrenheit
      - Gpm
      - WattHours
      - AmpsRms
      - VoltsRms
      - Gallons
      - ThermostatStateEnum

    For more information:
      - [ASLs](https://gridworks-type-registry.readthedocs.io/en/latest/)
      - [Global Authority](https://gridworks-type-registry.readthedocs.io/en/latest/enums.html#spaceheatunit)
    """

    Unknown = auto()
    Unitless = auto()
    W = auto()
    Celcius = auto()
    Fahrenheit = auto()
    Gpm = auto()
    WattHours = auto()
    AmpsRms = auto()
    VoltsRms = auto()
    Gallons = auto()
    ThermostatStateEnum = auto()

    @classmethod
    def default(cls) -> "Unit":
        return cls.Unknown

    @classmethod
    def values(cls) -> list[str]:
        return [elt.value for elt in cls]

    @classmethod
    def enum_name(cls) -> str:
        return "spaceheat.unit"

    @classmethod
    def enum_version(cls) -> str:
        return "001"



"""(Mostly) static functions describing ShNodes that were in Actor/ActorBase Scada/ScadaBase.

This will probably be refactored as we implement our local registry. Currently separated out here for clarity
because content is static (except for needing a path to the houses.json file, which we should be able to do
away with).
"""

import copy
import json
import typing
from collections import Counter, defaultdict
from dataclasses import dataclass
from functools import cached_property
from pathlib import Path
from typing import Any, Optional, TypeVar

from gw.errors import DcError

import gwproto.data_classes.components
from gwproto.data_classes.components import Ads111xBasedComponent, Component
from gwproto.data_classes.components.component import ComponentOnly
from gwproto.data_classes.components.electric_meter_component import (
    ElectricMeterComponent,
)
from gwproto.data_classes.data_channel import DataChannel
from gwproto.data_classes.resolver import ComponentResolver
from gwproto.data_classes.sh_node import ShNode
from gwproto.data_classes.synth_channel import SynthChannel
from gwproto.data_classes.telemetry_tuple import TelemetryTuple
from gwproto.decoders import (
    CacDecoder,
    ComponentDecoder,
)
from gwproto.default_decoders import (
    default_cac_decoder,
    default_component_decoder,
)
from gwproto.enums import ActorClass, TelemetryName
from gwproto.named_types import (
    ComponentAttributeClassGt,
    ComponentGt,
    DataChannelGt,
    ElectricMeterCacGt,
    SpaceheatNodeGt,
)

T = TypeVar("T")


@dataclass
class LoadError:
    type_name: str
    src_dict: dict[Any, Any]
    exception: Exception


class LoadArgs(typing.TypedDict):
    cacs: dict[str, ComponentAttributeClassGt]
    components: dict[str, Component[Any, Any]]
    nodes: dict[str, ShNode]
    data_channels: dict[str, DataChannel]
    synth_channels: dict[str, SynthChannel]


class HardwareLayout:
    layout: dict[Any, Any]
    cacs: dict[str, ComponentAttributeClassGt]
    components: dict[str, Component[Any, Any]]
    components_by_type: dict[type[Any], list[Component[Any, Any]]]
    nodes: dict[str, ShNode]
    nodes_by_component: dict[str, str]
    data_channels: dict[str, DataChannel]
    synth_channels: dict[str, SynthChannel]

    GT_SUFFIX = "Gt"

    @classmethod
    def load_cacs(
        cls,
        layout: dict[str, Any],
        *,
        raise_errors: bool = True,
        errors: Optional[list[LoadError]] = None,
        cac_decoder: Optional[CacDecoder] = None,
    ) -> dict[str, ComponentAttributeClassGt]:
        if errors is None:
            errors = []
        if cac_decoder is None:
            cac_decoder = default_cac_decoder
        cacs: dict[str, ComponentAttributeClassGt] = {}
        for type_name in [
            "Ads111xBasedCacs",
            "ResistiveHeaterCacs",
            "ElectricMeterCacs",
            "OtherCacs",
        ]:
            for cac_dict in layout.get(type_name, ()):
                try:
                    cac = cac_decoder.decode(cac_dict)
                    cacs[cac.ComponentAttributeClassId] = cac
                except Exception as e:  # noqa: PERF203
                    if raise_errors:
                        raise
                    errors.append(LoadError(type_name, cac_dict, e))
        return cacs

    @classmethod
    def get_data_class_name(cls, component_gt: ComponentGt) -> str:
        gt_class_name = component_gt.__class__.__name__
        if not gt_class_name.endswith(cls.GT_SUFFIX) or len(gt_class_name) <= len(
            cls.GT_SUFFIX
        ):
            raise DcError(  # noqa: TRY301
                f"Name of decoded component class ({gt_class_name}) "
                f"must end with <{cls.GT_SUFFIX}> "
                f"and be longer than {len(cls.GT_SUFFIX)} chars"
            )
        return gt_class_name[: -len(cls.GT_SUFFIX)]

    @classmethod
    def get_data_class_class(
        cls, component_gt: ComponentGt
    ) -> type[Component[Any, Any]]:
        return getattr(
            gwproto.data_classes.components,
            cls.get_data_class_name(component_gt),
            ComponentOnly,
        )

    @classmethod
    def make_component(
        cls, component_gt: ComponentGt, cac: ComponentAttributeClassGt
    ) -> Component[Any, Any]:
        return cls.get_data_class_class(component_gt)(gt=component_gt, cac=cac)

    @classmethod
    def load_components(
        cls,
        layout: dict[Any, Any],
        cacs: dict[str, ComponentAttributeClassGt],
        *,
        raise_errors: bool = True,
        errors: Optional[list[LoadError]] = None,
        component_decoder: Optional[ComponentDecoder] = None,
    ) -> dict[str, Component[Any, Any]]:
        if errors is None:
            errors = []
        if component_decoder is None:
            component_decoder = default_component_decoder
        components = {}
        for type_name in [
            "Ads111xBasedComponents",
            "ResistiveHeaterComponents",
            "ElectricMeterComponents",
            "OtherComponents",
        ]:
            for component_dict in layout.get(type_name, ()):
                try:
                    component_gt = component_decoder.decode(component_dict)
                    cac_gt = cacs.get(component_gt.ComponentAttributeClassId, None)
                    if cac_gt is None:
                        raise DcError(  # noqa: TRY301
                            f"cac {component_gt.ComponentAttributeClassId} not loaded for component "
                            f"<{component_gt.ComponentId}/<{component_gt.DisplayName}>\n"
                        )
                    components[component_gt.ComponentId] = cls.make_component(
                        component_gt,
                        cac_gt,
                    )
                except Exception as e:  # noqa: PERF203
                    if raise_errors:
                        raise
                    errors.append(LoadError(type_name, component_dict, e))
        return components

    @classmethod
    def make_node(
        cls,
        node_dict: dict[str, Any] | SpaceheatNodeGt,
        components: dict[str, Component[Any, Any]],
    ) -> ShNode:
        if isinstance(node_dict, SpaceheatNodeGt):
            node_gt = node_dict
        else:
            node_gt = SpaceheatNodeGt.model_validate(node_dict)
        if node_gt.ComponentId:
            component = components.get(node_gt.ComponentId)
            if component is None:
                raise ValueError(
                    f"ERROR. Component <{node_gt.ComponentId}> not loaded "
                    f"for node <{node_gt.Name}>"
                )
        else:
            component = None
        return ShNode(component=component, **node_gt.model_dump())

    @classmethod
    def load_nodes(
        cls,
        layout: dict[Any, Any],
        components: dict[str, Component[Any, Any]],
        *,
        raise_errors: bool = True,
        errors: Optional[list[LoadError]] = None,
        included_node_names: Optional[set[str]] = None,
    ) -> dict[str, ShNode]:
        nodes = {}
        if errors is None:
            errors = []
        for node_dict in layout.get("ShNodes", []):
            try:
                node_name = node_dict["Name"]
                if included_node_names is None or node_name in included_node_names:
                    nodes[node_name] = cls.make_node(node_dict, components)
            except Exception as e:  # noqa: PERF203
                if raise_errors:
                    raise
                errors.append(LoadError("ShNode", node_dict, e))
        return nodes

    @classmethod
    def make_channel(
        cls, dc_dict: dict[str, Any], nodes: dict[str, ShNode]
    ) -> DataChannel:
        data_channel_gt = DataChannelGt.model_validate(dc_dict)
        about_node = nodes.get(data_channel_gt.AboutNodeName)
        captured_by_node = nodes.get(data_channel_gt.CapturedByNodeName)
        if about_node is None or captured_by_node is None:
            raise ValueError(
                f"ERROR. DataChannel related nodes must exist for {dc_dict.get('Name')}!\n"
                f"  For AboutNodeName <{data_channel_gt.AboutNodeName}> "
                f"got {about_node}\n"
                f"  for CapturedByNodeName <{data_channel_gt.CapturedByNodeName}>"
                f"got {captured_by_node}"
            )
        return DataChannel(
            about_node=about_node, captured_by_node=captured_by_node, **dc_dict
        )

    @classmethod
    def make_synth_channel(
        cls, synth_dict: dict[str, Any], nodes: dict[str, ShNode]
    ) -> SynthChannel:
        created_by_node_name = synth_dict.get("CreatedByNodeName", "")
        created_by_node = nodes.get(created_by_node_name)
        if created_by_node is None:
            raise ValueError(
                f"ERROR. SynthChannel related nodes must exist for {synth_dict.get('Name')}!\n"
                f"  For CreatedByNodeName<{created_by_node_name}> "
                f"got None!\n"
            )
        return SynthChannel(created_by_node=created_by_node, **synth_dict)

    @classmethod
    def check_dc_id_uniqueness(
        cls,
        data_channels: dict[str, DataChannel],
    ) -> None:
        id_counter = Counter(dc.Id for dc in data_channels.values())
        dupes = [node_id for node_id, count in id_counter.items() if count > 1]
        if dupes:
            raise DcError(f"Duplicate dc.Id(s) found: {dupes}")

    @classmethod
    def check_node_channel_consistency(
        cls, nodes: dict[str, ShNode], data_channels: dict[str, DataChannel]
    ) -> None:
        capturing_classes = [
            ActorClass.PowerMeter,
            ActorClass.MultipurposeSensor,
        ]
        active_nodes = [
            node for node in nodes.values() if node.ActorClass in capturing_classes
        ]
        for node in active_nodes:
            if node.component is None:
                my_channel_names = []
            else:
                my_channel_names = [
                    config.ChannelName for config in node.component.gt.ConfigList
                ]
            my_channels = [
                dc for dc in data_channels.values() if dc.Name in my_channel_names
            ]
            for channel in my_channels:
                if channel.CapturedByNodeName != node.Name:
                    raise DcError(
                        f"Channel {channel} should have CapturedByNodeName {node.Name}"
                    )

    @classmethod
    def check_data_channel_consistency(
        cls,
        nodes: dict[str, ShNode],
        components: dict[str, Component[Any, Any]],
        data_channels: dict[str, DataChannel],
    ) -> None:
        cls.check_dc_id_uniqueness(data_channels)
        dc_names_by_component: set[str] = set()
        for c in components.values():
            channel_names = {config.ChannelName for config in c.gt.ConfigList}
            if dc_names_by_component & channel_names:
                raise DcError(
                    f"Channel name overlap!: {dc_names_by_component & channel_names}"
                )
            dc_names_by_component.update(channel_names)
        actual_dc_names = {dc.Name for dc in data_channels.values()}
        if dc_names_by_component != actual_dc_names:
            by_comp = list(dc_names_by_component)
            by_comp.sort()
            actual = list(actual_dc_names)
            actual.sort()
            raise DcError(
                "Channel inconsistency! \n"
                f"From Components:{by_comp}\n"
                f"From DataChannel list:{actual}\n"
            )
        cls.check_node_channel_consistency(nodes, data_channels)

    @classmethod
    def check_actor_component_consistency(cls, nodes: dict[str, ShNode]) -> None:
        pm_nodes = [
            node for node in nodes.values() if node.ActorClass == ActorClass.PowerMeter
        ]
        for node in pm_nodes:
            if (
                node.component is None
                or node.component.gt.TypeName != "electric.meter.component.gt"
            ):
                raise DcError(
                    f"Power Meter node {node} needs ElectricMeterComponent."
                    f"Got component {node.component}"
                )
        em_nodes = [
            node
            for node in nodes.values()
            if node.ActorClass == ActorClass.MultipurposeSensor
        ]
        for node in em_nodes:
            multi_comp_type_names = ["ads111x.based.component.gt"]
            if (
                node.component is None
                or node.component.gt.TypeName not in multi_comp_type_names
            ):
                raise DcError(
                    f"Power Meter node {node} needs Component "
                    f"in {multi_comp_type_names}. Got component"
                    f"{node.component}"
                )

    @classmethod
    def check_handle_hierarchy(cls, nodes: dict[str, ShNode]) -> None:
        for n in nodes.values():
            boss_handle = cls.boss_handle(n.handle)
            # No dots in your name: you are your own boss
            if boss_handle:
                boss = next(
                    (n for n in nodes.values() if n.handle == boss_handle), None
                )
                if boss is None:
                    raise DcError(f"{n.name} is missing boss {boss_handle}")

    @classmethod
    def check_node_unique_ids(cls, nodes: dict[str, ShNode]) -> None:
        id_counter = Counter(node.ShNodeId for node in nodes.values())
        dupes = [node_id for node_id, count in id_counter.items() if count > 1]
        if dupes:
            raise DcError(f"Duplicate ShNodeId(s) found: {dupes}")

    @classmethod
    def check_transactive_metering_consistency(
        cls, nodes: dict[str, ShNode], data_channels: dict[str, DataChannel]
    ) -> None:
        transactive_nodes = {node for node in nodes.values() if node.InPowerMetering}
        transactive_channels = {
            dc for dc in data_channels.values() if dc.InPowerMetering
        }
        # Part 1: If a data channel is in transactive_channels, its about_node must be in transactive_nodes
        for tc in transactive_channels:
            if tc.about_node not in transactive_nodes:
                raise DcError(
                    f"Data channel {tc} has about_node {tc.about_node}, which does not have InPowerMetering!"
                )
        # Check condition 2: If a node is in transactive_nodes, there must be a data channel with that node as about_node
        for node in transactive_nodes:
            if not any(tc for tc in transactive_channels if tc.about_node == node):
                raise DcError(
                    f"Node {node} is in transactive_nodes but no data channel with InPowerMetering has this node as about_node"
                )

    @classmethod
    def check_ads_terminal_block_consistency(cls, c: Ads111xBasedComponent) -> None:
        possible_indices = set(
            range(1, c.cac.TotalTerminalBlocks + 1)
        )  # e,g {1, .., 12}
        actual_indices = {tc.TerminalBlockIdx for tc in c.gt.ConfigList}
        if not actual_indices.issubset(possible_indices):
            raise DcError(
                f"Terminal Block indices {actual_indices}"
                f"When Ads only has {c.cac.TotalTerminalBlocks} terminal blocks!"
            )

    @classmethod
    def load_data_channels(
        cls,
        layout: dict[Any, Any],
        nodes: dict[str, ShNode],
        *,
        raise_errors: bool = True,
        errors: Optional[list[LoadError]] = None,
    ) -> dict[str, DataChannel]:
        dcs = {}
        if errors is None:
            errors = []
        for dc_dict in layout.get("DataChannels", []):
            try:
                dc_name = dc_dict["Name"]
                dcs[dc_name] = cls.make_channel(dc_dict, nodes)
            except Exception as e:  # noqa: PERF203
                if raise_errors:
                    raise
                errors.append(LoadError("DataChannel", dc_dict, e))
        return dcs

    @classmethod
    def load_synth_channels(
        cls,
        layout: dict[Any, Any],
        nodes: dict[str, ShNode],
        *,
        raise_errors: bool = True,
        errors: Optional[list[LoadError]] = None,
    ) -> dict[str, SynthChannel]:
        synths = {}
        if errors is None:
            errors = []
        for d in layout.get("SynthChannels", []):
            try:
                name = d["Name"]
                synths[name] = cls.make_synth_channel(d, nodes)
            except Exception as e:  # noqa: PERF203
                if raise_errors:
                    raise
                errors.append(LoadError("SynthChannel", d, e))
        return synths

    @classmethod
    def resolve_node_links(
        cls,
        node: ShNode,
        all_nodes: dict[str, ShNode],
        components: dict[str, Component[Any, Any]],
        *,
        raise_errors: bool = True,
        errors: Optional[list[LoadError]] = None,
    ) -> None:
        if errors is None:
            errors = []
        try:
            if node.component_id is not None:
                component = components.get(node.component_id, None)
                if component is None:
                    raise DcError(  # noqa: TRY301
                        f"{node.name} component {node.component_id} not loaded!"
                    )
                if isinstance(component, ComponentResolver):
                    component.resolve(
                        node.name,
                        all_nodes,
                        components,
                    )
        except Exception as e:
            if raise_errors:
                raise
            errors.append(
                LoadError("ShNode", {"node": {"name": node.Name, "node": node}}, e)
            )

    @classmethod
    def resolve_links(
        cls,
        nodes: dict[str, ShNode],
        components: dict[str, Component[Any, Any]],
        *,
        raise_errors: bool = True,
        errors: Optional[list[LoadError]] = None,
    ) -> None:
        if errors is None:
            errors = []
        for node in nodes.values():
            cls.resolve_node_links(
                node=node,
                all_nodes=nodes,
                components=components,
                raise_errors=raise_errors,
                errors=errors,
            )

    def __init__(  # noqa: PLR0913
        self,
        layout: dict[Any, Any],
        *,
        cacs: dict[str, ComponentAttributeClassGt],
        components: dict[str, Component[Any, Any]],
        nodes: dict[str, ShNode],
        data_channels: dict[str, DataChannel],
        synth_channels: dict[str, SynthChannel],
    ) -> None:
        self.layout = copy.deepcopy(layout)
        self.cacs = dict(cacs)
        self.components = dict(components)
        self.components_by_type = defaultdict(list)
        for component in self.components.values():
            self.components_by_type[type(component)].append(component)
        self.nodes = dict(nodes)
        self.nodes_by_component = {
            node.component_id: node.name
            for node in self.nodes.values()
            if node.component_id is not None
        }
        self.data_channels = dict(data_channels)
        self.synth_channels = dict(synth_channels)

    def clear_property_cache(self) -> None:
        for cached_prop_name in [
            prop_name
            for prop_name in type(self).__dict__
            if isinstance(type(self).__dict__[prop_name], cached_property)
        ]:
            self.__dict__.pop(cached_prop_name, None)

    @classmethod
    def load(  # noqa: PLR0913
        cls,
        layout_path: Path | str,
        *,
        included_node_names: Optional[set[str]] = None,
        raise_errors: bool = True,
        errors: Optional[list[LoadError]] = None,
        cac_decoder: Optional[CacDecoder] = None,
        component_decoder: Optional[ComponentDecoder] = None,
    ) -> "HardwareLayout":
        with Path(layout_path).open() as f:
            layout = json.loads(f.read())
        return cls.load_dict(
            layout,
            included_node_names=included_node_names,
            raise_errors=raise_errors,
            errors=errors,
            cac_decoder=cac_decoder,
            component_decoder=component_decoder,
        )

    @classmethod
    def validate_layout(  # noqa: C901
        cls,
        load_args: LoadArgs,
        *,
        raise_errors: bool,
        errors: Optional[list[LoadError]] = None,
    ) -> None:
        nodes = load_args["nodes"]
        components = load_args["components"]
        data_channels = load_args["data_channels"]
        errors_caught = []
        try:
            cls.check_node_unique_ids(nodes)
        except Exception as e:  # noqa: BLE001
            errors_caught.append(LoadError("hardware.layout", nodes, e))
        try:
            cls.check_handle_hierarchy(nodes)
        except Exception as e:
            if raise_errors:
                raise
            errors_caught.append(LoadError("hardware.layout", nodes, e))
        try:
            cls.check_actor_component_consistency(nodes)
        except Exception as e:  # noqa: BLE001
            errors_caught.append(LoadError("hardware.layout", nodes, e))
        try:
            cls.check_data_channel_consistency(
                nodes,
                components,
                data_channels,
            )
            cls.check_transactive_metering_consistency(
                nodes,
                data_channels,
            )
        except Exception as e:  # noqa: BLE001
            errors_caught.append(LoadError("data.channel.gt", data_channels, e))
        ads111x_components = [
            comp
            for comp in components.values()
            if isinstance(comp, Ads111xBasedComponent)
        ]
        for c in ads111x_components:
            try:
                cls.check_ads_terminal_block_consistency(c)
            except Exception as e:  # noqa: BLE001, PERF203
                errors_caught.append(
                    LoadError("ads111x.based.component.gt", c.gt.model_dump(), e)
                )
        if errors_caught:
            if raise_errors:
                s = "ERROR in HardwareLayout validation. Caught:\n"
                for error in errors_caught:
                    s += f"  TypeName: {error.type_name}  Exception: <{error.exception}>  src: {error.src_dict}\n"
                raise DcError(s)
            if errors is not None:
                errors.extend(errors_caught)

    @classmethod
    def load_dict(  # noqa: PLR0913
        cls,
        layout: dict[Any, Any],
        *,
        included_node_names: Optional[set[str]] = None,
        raise_errors: bool = True,
        errors: Optional[list[LoadError]] = None,
        cac_decoder: Optional[CacDecoder] = None,
        component_decoder: Optional[ComponentDecoder] = None,
    ) -> "HardwareLayout":
        if errors is None:
            errors = []
        cacs = cls.load_cacs(
            layout=layout,
            raise_errors=raise_errors,
            errors=errors,
            cac_decoder=cac_decoder,
        )
        components = cls.load_components(
            layout=layout,
            cacs=cacs,
            raise_errors=raise_errors,
            errors=errors,
            component_decoder=component_decoder,
        )
        nodes = cls.load_nodes(
            layout=layout,
            components=components,
            raise_errors=raise_errors,
            errors=errors,
            included_node_names=included_node_names,
        )
        data_channels = cls.load_data_channels(
            layout=layout,
            nodes=nodes,
            raise_errors=raise_errors,
            errors=errors,
        )
        synth_channels = cls.load_synth_channels(
            layout=layout,
            nodes=nodes,
            raise_errors=raise_errors,
            errors=errors,
        )
        load_args: LoadArgs = {
            "cacs": cacs,
            "components": components,
            "nodes": nodes,
            "data_channels": data_channels,
            "synth_channels": synth_channels,
        }
        cls.resolve_links(
            load_args["nodes"],
            load_args["components"],
            raise_errors=raise_errors,
            errors=errors,
        )
        cls.validate_layout(load_args, raise_errors=raise_errors, errors=errors)
        return HardwareLayout(layout, **load_args)

    def add_node(self, node: dict[str, Any] | SpaceheatNodeGt) -> ShNode:
        node = self.make_node(node, self.components)
        if node.Name in self.nodes:
            raise ValueError(f"ERROR. Node with name {node.Name} already exists")
        if node.ComponentId in self.nodes_by_component:
            raise ValueError(
                f"ERROR. Node with component id {node.ComponentId} "
                "already exists. "
                f"Tried to add node {node.Name}. Existing node is "
                f"{self.nodes_by_component[node.ComponentId]}"
            )
        self.nodes[node.Name] = node
        self.resolve_node_links(node, self.nodes, self.components, raise_errors=True)
        if node.ComponentId is not None:
            self.nodes_by_component[node.ComponentId] = node.Name
        self.clear_property_cache()
        return node

    def channel(self, name: str, default: Any = None) -> DataChannel:  # noqa: ANN401
        return self.data_channels.get(name, default)

    def synth_channel(self, name: str, default: Any = None) -> SynthChannel:  # noqa: ANN401
        return self.synth_channels.get(name, default)

    def node(self, name: str, default: Any = None) -> ShNode:  # noqa: ANN401
        return self.nodes.get(name, default)

    def node_by_handle(self, handle: str) -> Optional[ShNode]:
        d = {node.Handle: node for node in self.nodes.values() if node.Handle}
        if handle in d:
            return d[handle]
        return None

    def component(self, node_name: str) -> Optional[Component[Any, Any]]:
        return self.component_from_node(self.node(node_name, None))

    def cac(self, node_name: str) -> Optional[ComponentAttributeClassGt]:
        component = self.component(node_name)
        if component is None:
            return None
        return typing.cast(ComponentAttributeClassGt, component.cac)

    def get_component_as_type(self, component_id: str, type_: type[T]) -> Optional[T]:
        component = self.components.get(component_id, None)
        if component is not None and not isinstance(component, type_):
            raise ValueError(
                f"ERROR. Component <{component_id}> has type {type(component)} not {type_}"
            )
        return component

    def get_components_by_type(self, type_: type[T]) -> list[T]:
        entries = self.components_by_type.get(type_, [])
        for i, entry in enumerate(entries):
            if not isinstance(entry, type_):
                raise TypeError(
                    f"ERROR. Entry {i + 1} in "
                    f"HardwareLayout.components_by_typ[{type_}] "
                    f"has the wrong type {type(entry)}"
                )
        return typing.cast(list[T], entries)

    def node_from_component(self, component_id: str) -> Optional[ShNode]:
        return self.nodes.get(self.nodes_by_component.get(component_id, ""), None)

    def component_from_node(
        self, node: Optional[ShNode]
    ) -> Optional[Component[Any, Any]]:
        return (
            self.components.get(
                node.component_id if node.component_id is not None else "", None
            )
            if node is not None
            else None
        )

    @classmethod
    def parent_hierarchy_name(cls, hierarchy_name: str) -> Optional[str]:
        last_delimiter = hierarchy_name.rfind(".")
        if last_delimiter == -1:
            return None
        return hierarchy_name[:last_delimiter]

    def parent_node(self, node: ShNode) -> Optional[ShNode]:
        h_name = self.parent_hierarchy_name(node.actor_hierarchy_name)
        if not h_name:
            return None
        parent = next(
            (n for n in self.nodes.values() if n.actor_hierarchy_name == h_name), None
        )
        if parent is None:
            raise DcError(f"{node} is missing parent {h_name}!")
        return self.node(h_name)

    @classmethod
    def boss_handle(cls, handle: str) -> Optional[str]:
        if "." not in handle:
            return None
        return ".".join(handle.split(".")[:-1])

    def boss_node(self, node: ShNode) -> Optional[ShNode]:
        boss_handle = self.boss_handle(node.handle)
        # No dots in your name: you are your own boss
        if not boss_handle:
            return node
        boss = next((n for n in self.nodes.values() if n.handle == boss_handle), None)
        if boss is None:
            raise DcError(f"{node} is missing boss {boss_handle}")
        return boss

    def direct_reports(self, node: ShNode) -> list[ShNode]:
        return [n for n in self.nodes.values() if self.boss_node(n) == node]

    def node_from_handle(self, handle: str) -> Optional[ShNode]:
        return next((n for n in self.nodes.values() if n.handle == handle), None)

    @cached_property
    def atn_g_node_alias(self) -> str:
        return self.layout["MyAtomicTNodeGNode"]["Alias"]  # type: ignore[no-any-return]

    @cached_property
    def atn_g_node_instance_id(self) -> str:
        return self.layout["MyAtomicTNodeGNode"]["GNodeId"]  # type: ignore[no-any-return]

    @cached_property
    def atn_g_node_id(self) -> str:
        return self.layout["MyAtomicTNodeGNode"]["GNodeId"]  # type: ignore[no-any-return]

    @cached_property
    def terminal_asset_g_node_alias(self) -> str:
        my_atn_as_dict = self.layout["MyTerminalAssetGNode"]
        return my_atn_as_dict["Alias"]  # type: ignore[no-any-return]

    @cached_property
    def terminal_asset_g_node_id(self) -> str:
        my_atn_as_dict = self.layout["MyTerminalAssetGNode"]
        return my_atn_as_dict["GNodeId"]  # type: ignore[no-any-return]

    @cached_property
    def scada_g_node_alias(self) -> str:
        my_scada_as_dict = self.layout["MyScadaGNode"]
        return my_scada_as_dict["Alias"]  # type: ignore[no-any-return]

    @cached_property
    def scada_g_node_id(self) -> str:
        my_scada_as_dict = self.layout["MyScadaGNode"]
        return my_scada_as_dict["GNodeId"]  # type: ignore[no-any-return]

    @cached_property
    def all_telemetry_tuples_for_agg_power_metering(self) -> list[TelemetryTuple]:
        telemetry_tuples = []
        for node in self.all_nodes_in_agg_power_metering:
            telemetry_tuples += [
                TelemetryTuple(
                    AboutNode=node,
                    SensorNode=self.power_meter_node,
                    TelemetryName=TelemetryName.PowerW,
                )
            ]
        return telemetry_tuples

    @cached_property
    def all_nodes_in_agg_power_metering(self) -> list[ShNode]:
        """All nodes whose power level is metered and included in power reporting by the Scada"""
        all_nodes = list(self.nodes.values())
        return list(filter(lambda x: x.in_power_metering, all_nodes))

    @cached_property
    def all_power_meter_telemetry_tuples(self) -> list[TelemetryTuple]:
        return [
            TelemetryTuple(
                AboutNode=self.nodes[
                    self.data_channels[config.ChannelName].AboutNodeName
                ],
                SensorNode=self.power_meter_node,
                TelemetryName=self.data_channels[config.ChannelName].TelemetryName,
            )
            for config in self.power_meter_component.gt.ConfigList
        ]

    @cached_property
    def power_meter_node(self) -> ShNode:
        return next(
            filter(lambda x: x.ActorClass == ActorClass.PowerMeter, self.nodes.values())
        )

    @cached_property
    def power_meter_component(self) -> ElectricMeterComponent:
        if self.power_meter_node.component is None:
            raise ValueError(
                f"ERROR. power_meter_node {self.power_meter_node} has no component."
            )
        return typing.cast(ElectricMeterComponent, self.power_meter_node.component)

    @cached_property
    def power_meter_cac(self) -> ElectricMeterCacGt:
        if isinstance(self.power_meter_component.cac, ElectricMeterCacGt):
            return self.power_meter_component.cac
        raise TypeError(
            f"ERROR. power_meter_component cac {self.power_meter_component.cac}"
            f" / {type(self.power_meter_component.cac)} is not an ElectricMeterCac"
        )

    @cached_property
    def all_multipurpose_telemetry_tuples(self) -> list[TelemetryTuple]:
        multi_nodes = list(
            filter(
                lambda x: (
                    x.actor_class
                    in {
                        ActorClass.MultipurposeSensor,
                        ActorClass.HubitatTankModule,
                        ActorClass.HubitatPoller,
                        ActorClass.HoneywellThermostat,
                    }
                ),
                self.nodes.values(),
            )
        )
        telemetry_tuples: list[TelemetryTuple] = []
        for node in multi_nodes:
            if node.component is not None and (
                channels := [
                    self.data_channels[cfg.ChannelName]
                    for cfg in node.component.gt.ConfigList
                ]
            ):
                telemetry_tuples.extend(
                    TelemetryTuple(
                        AboutNode=ch.about_node,
                        SensorNode=ch.captured_by_node,
                        TelemetryName=ch.TelemetryName,
                    )
                    for ch in channels
                )
        return telemetry_tuples

    @cached_property
    def my_telemetry_tuples(self) -> list[TelemetryTuple]:
        """This will include telemetry tuples from all the multipurpose sensors, the most
        important of which is the power meter."""
        return (
            self.all_power_meter_telemetry_tuples
            + self.all_multipurpose_telemetry_tuples
        )



"""Type dfr.config, version 000"""

from typing import Literal

from pydantic import PositiveInt, StrictInt

from gwproto.named_types import ChannelConfig


class DfrConfig(ChannelConfig):
    OutputIdx: PositiveInt
    InitialVoltsTimes100: StrictInt
    TypeName: Literal["dfr.config"] = "dfr.config"
    Version: str = "000"



"""Type ticklist.hall.report, version 000"""

from typing import Literal

from pydantic import BaseModel

from gwproto.named_types.ticklist_hall import TicklistHall
from gwproto.property_format import LeftRightDotStr, SpaceheatName, UTCMilliseconds


class TicklistHallReport(BaseModel):
    """
    Used by the SCADA to forward a ticklist.hall message received from a PicoFlowHall module.
    """

    TerminalAssetAlias: LeftRightDotStr
    ChannelName: SpaceheatName
    ScadaReceivedUnixMs: UTCMilliseconds
    Ticklist: TicklistHall
    TypeName: Literal["ticklist.hall.report"] = "ticklist.hall.report"
    Version: str = "000"



from abc import ABC, abstractmethod
from typing import Any

from gwproto.data_classes.components.component import Component
from gwproto.data_classes.sh_node import ShNode


class ComponentResolver(ABC):
    @abstractmethod
    def resolve(
        self,
        node_name: str,
        nodes: dict[str, ShNode],
        components: dict[str, Component[Any, Any]],
    ) -> None:
        raise NotImplementedError



"""ShNode definition"""

from typing import Any, Optional

from pydantic import ConfigDict

from gwproto.data_classes.components.component import Component
from gwproto.enums import ActorClass as ActorClassEnum
from gwproto.named_types import SpaceheatNodeGt


def parent_hierarchy_name(hierarchy_name: str) -> str:
    last_delimiter = hierarchy_name.rfind(".")
    if last_delimiter == -1:
        return hierarchy_name
    return hierarchy_name[:last_delimiter]


class ShNode(SpaceheatNodeGt):
    """
    A SpaceheatNode, or ShNode, is an organizing principal for the SCADA software.
    ShNodes can represent both underlying physical objects (water tank), measurements of these
    objects (temperature sensing at the top of a water tank), and actors within the code
    (an actor measuring multiple temperatures, or an actor responsible for filtering/smoothing
    temperature data for the purposes of thermostatic control).
    """

    component: Optional[Component[Any, Any]] = None
    model_config = ConfigDict(arbitrary_types_allowed=True)

    def __hash__(self) -> int:
        return hash(self.ShNodeId)

    @property
    def sh_node_id(self) -> str:
        return self.ShNodeId

    @property
    def name(self) -> str:
        return self.Name

    @property
    def actor_hierarchy_name(self) -> str:
        if self.ActorHierarchyName is None:
            return self.Name
        return self.ActorHierarchyName

    @property
    def handle(self) -> str:
        if self.Handle is None:
            return self.Name
        return self.Handle

    @property
    def actor_class(self) -> ActorClassEnum:
        return ActorClassEnum(self.ActorClass)

    @property
    def actor_class_str(self) -> str:
        return self.ActorClass

    @property
    def display_name(self) -> Optional[str]:
        return self.DisplayName

    @property
    def component_id(self) -> Optional[str]:
        return self.ComponentId

    @property
    def in_power_metering(self) -> Optional[bool]:
        return self.InPowerMetering

    def __repr__(self) -> str:
        rs = f"ShNode {self.display_name} => {self.name}, "
        if self.has_actor:
            rs += f" ({self.actor_class})"
        else:
            rs += " (passive, no actor)"
        return rs

    @property
    def has_actor(self) -> bool:
        return self.actor_class != ActorClassEnum.NoActor

    def to_gt(self) -> SpaceheatNodeGt:
        # Copy the current instance excluding the extra fields
        return SpaceheatNodeGt(**self.model_dump(exclude={"component"}))



"""Type electric.meter.cac.gt, version 010"""

from typing import Literal, Optional

from gwproto.enums import TelemetryName
from gwproto.named_types import ComponentAttributeClassGt


class ElectricMeterCacGt(ComponentAttributeClassGt):
    TelemetryNameList: list[TelemetryName]
    DefaultBaud: Optional[int] = None
    TypeName: Literal["electric.meter.cac.gt"] = "electric.meter.cac.gt"
    Version: str = "001"



# Literal Enum:
#  - no additional values can be added over time.
#  - Sent as-is, not in hex symbol
from enum import auto

from gw.enums import GwStrEnum


class RelayClosedOrOpen(GwStrEnum):
    """
    These are fsm states (as opposed to readings from a pin).
    """

    RelayClosed = auto()
    RelayOpen = auto()

    @classmethod
    def values(cls) -> list[str]:
        """
        Returns enum choices
        """
        return [elt.value for elt in cls]

    @classmethod
    def default(cls) -> "RelayClosedOrOpen":
        return cls.RelayClosed

    @classmethod
    def enum_name(cls) -> str:
        return "relay.closed.or.open"



from enum import auto

from gw.enums import GwStrEnum


class FsmReportType(GwStrEnum):
    """

    Values:
      - Other
      - Event
      - Action

    For more information:
      - [ASLs](https://gridworks-type-registry.readthedocs.io/en/latest/)
      - [Global Authority](https://gridworks-type-registry.readthedocs.io/en/latest/enums.html#fsmreporttype)
      - [More Info](https://gridworks-protocol.readthedocs.io/en/latest/finite-state-machines.html)
    """

    Other = auto()
    Event = auto()
    Action = auto()

    @classmethod
    def default(cls) -> "FsmReportType":
        return cls.Other

    @classmethod
    def values(cls) -> list[str]:
        return [elt.value for elt in cls]

    @classmethod
    def enum_name(cls) -> str:
        return "fsm.report.type"

    @classmethod
    def enum_version(cls) -> str:
        return "000"



# Literal Enum:
#  - no additional values can be added over time.
#  - Sent as-is, not in hex symbol
from enum import auto

from gwproto.enums.relay_event_base import RelayEventBase


class ChangeRelayState(RelayEventBase):
    """
    This is meant for relays in either NormallyOpen or NormallyClosed configuration (i.e. not
    configured as DoubleThrow). It provides the natural Finite State Machine actions that go
    along with simple.relay.state states of RelayOpen and RelayClosed. Open = PowerOff = 0 and
    Closed = PowerOn = 1
    """

    CloseRelay = auto()
    OpenRelay = auto()

    @classmethod
    def values(cls) -> list[str]:
        """
        Returns enum choices
        """
        return [elt.value for elt in cls]

    @classmethod
    def default(cls) -> "ChangeRelayState":
        return cls.OpenRelay

    @classmethod
    def enum_name(cls) -> str:
        return "change.relay.state"



"""Type dfr.component.gt, version 000"""

from collections.abc import Sequence
from typing import Literal

from pydantic import PositiveInt

from gwproto.named_types import ComponentGt
from gwproto.named_types.dfr_config import DfrConfig


class DfrComponentGt(ComponentGt):
    ConfigList: Sequence[DfrConfig]
    I2cAddressList: list[PositiveInt]
    TypeName: Literal["dfr.component.gt"] = "dfr.component.gt"
    Version: str = "000"



"""PicoTankModuleComponent definition"""

from gwproto.data_classes.components.component import Component
from gwproto.named_types import ComponentAttributeClassGt, PicoTankModuleComponentGt


class PicoTankModuleComponent(
    Component[PicoTankModuleComponentGt, ComponentAttributeClassGt]
): ...



from typing import NamedTuple

from pydantic import BaseModel

from gwproto.data_classes.sh_node import ShNode
from gwproto.enums import TelemetryName
from gwproto.property_format import SpaceheatName


class TelemetryTuple(NamedTuple):
    AboutNode: ShNode
    SensorNode: ShNode
    TelemetryName: TelemetryName

    def __repr__(self) -> str:
        return f"TT({self.AboutNode.name} {self.TelemetryName} read by {self.SensorNode.name})"


class ChannelStub(BaseModel):
    Name: SpaceheatName
    AboutNodeName: SpaceheatName
    TelemetryName: TelemetryName
    InPowerMetering: bool = False

    def __hash__(self) -> int:
        return hash(self.Name)



# ruff: noqa: ANN401
import re
import uuid
from datetime import datetime, timezone
from typing import Annotated

from gw.enums import MarketTypeName
from pydantic import BeforeValidator, Field

UTC_2000_01_01_TIMESTAMP = datetime(2000, 1, 1, tzinfo=timezone.utc).timestamp()
UTC_3000_01_01_TIMESTAMP = datetime(3000, 1, 1, tzinfo=timezone.utc).timestamp()


def check_is_log_style_date_with_millis(v: str) -> None:
    """Checks LogStyleDateWithMillis format

    LogStyleDateWithMillis format:  YYYY-MM-DDTHH:mm:ss.SSS

    Args:
        v (str): the candidate

    Raises:
        ValueError: if v is not LogStyleDateWithMillis format.
        In particular the milliseconds must have exactly 3 digits.
    """
    correct_millisecond_part_length = 3
    try:
        datetime.fromisoformat(v)
    except ValueError as e:
        raise ValueError(f"{v} is not in LogStyleDateWithMillis format") from e
    # The python fromisoformat allows for either 3 digits (milli) or 6 (micro)
    # after the final period. Make sure its 3
    milliseconds_part = v.split(".")[1]
    if len(milliseconds_part) != correct_millisecond_part_length:
        raise ValueError(
            f"{v} is not in LogStyleDateWithMillis format."
            " Milliseconds must have exactly 3 digits"
        )


def is_handle_name(v: str) -> str:
    """
    HandleName format: words separated by periods, where the worlds are lowercase
    alphanumeric plus hyphens
    """
    try:
        x = v.split(".")
    except Exception as e:
        raise ValueError(f"Failed to seperate <{v}> into words with split'.'") from e
    first_word = x[0]
    first_char = first_word[0]
    if not first_char.isalpha():
        raise ValueError(
            f"Most significant word of <{v}> must start wif64th alphabet char."
        )
    for word in x:
        for char in word:
            if not (char.isalnum() or char == "-"):
                raise ValueError(
                    f"words of <{v}> split by by '.' must be alphanumeric or hyphen."
                )
    if not v.islower():
        raise ValueError(f" <{v}> must be lowercase.")
    return v


def is_hex_char(v: str) -> str:
    """Checks HexChar format

    HexChar format: single-char string in '0123456789abcdefABCDEF'

    Args:
        v (str): the candidate

    Raises:
        ValueError: if v is not HexChar format
    """
    if not isinstance(v, str):
        raise ValueError(f"<{v}> must be string. Got type <{type(v)}")  # noqa: TRY004
    if len(v) > 1:
        raise ValueError(f"<{v}> must be a hex char, but not of len 1")
    if v not in "0123456789abcdefABCDEF":
        raise ValueError(f"<{v}> must be one of '0123456789abcdefABCDEF'")
    return v


def is_int(v: int) -> int:
    if not isinstance(v, int):
        raise TypeError("Not an integer!")
    return v


def is_left_right_dot(candidate: str) -> str:
    """Lowercase AlphanumericStrings separated by dots (i.e. periods), with most
    significant word to the left.  I.e. `d1.ne` is the child of `d1`.
    Checking the format cannot verify the significance of words. All
    words must be alphanumeric. Most significant word must start with
    an alphabet charecter


    Raises:
        ValueError: if candidate is not of lrd format (e.g. d1.iso.me.apple)
    """
    try:
        x: list[str] = candidate.split(".")
    except Exception as e:
        raise ValueError("Failed to seperate into words with split'.'") from e
    first_word = x[0]
    first_char = first_word[0]
    if not first_char.isalpha():
        raise ValueError(
            f"Most significant word must start with alphabet char. Got '{first_word}'"
        )
    for word in x:
        if not word.isalnum():
            raise ValueError(
                f"words seperated by dots must be alphanumeric. Got '{word}'"
            )
    if not candidate.islower():
        raise ValueError(f"alias must be lowercase. Got '{candidate}'")
    return candidate


MAC_REGEX = re.compile("[0-9a-f]{2}([-:]?)[0-9a-f]{2}(\\1[0-9a-f]{2}){4}$")


def has_mac_address_format(mac_str: str) -> bool:
    return bool(MAC_REGEX.match(mac_str.lower()))


def is_spaceheat_name(v: str) -> str:
    """
    SpaceheatName format: Lowercase alphanumeric words separated by hypens
    """
    try:
        x = v.split("-")
    except Exception as e:
        raise ValueError(
            f"<{v}>: Fails SpaceheatName format! Failed to seperate into words with split'-'"
        ) from e
    first_word = x[0]
    first_char = first_word[0]
    if not first_char.isalpha():
        raise ValueError(
            f"<{v}>: Fails SpaceheatName format! Most significant word  must start with alphabet char."
        )
    for word in x:
        if not word.isalnum():
            raise ValueError(
                f"<{v}>: Fails SpaceheatName format! words of split by by '-' must be alphanumeric."
            )
    if not v.islower():
        raise ValueError(
            f"<{v}>: Fails SpaceheatName format! All characters of  must be lowercase."
        )
    return v


def is_uuid4_str(v: str) -> str:
    v = str(v)
    try:
        u = uuid.UUID(v)
    except Exception as e:
        raise ValueError(f"Invalid UUID4: <{v}  <{e}>") from e
    if u.version != 4:
        raise ValueError(f"{v} is valid uid, but of version {u.version}, not 4")
    return str(u)


def is_world_instance_name_format(candidate: str) -> bool:
    try:
        words = candidate.split("__")
    except:  # noqa
        return False
    if len(words) != 2:
        return False
    try:
        int(words[1])
    except:  # noqa
        return False
    try:
        root_g_node_alias_words = words[0].split(".")
    except:  # noqa
        return False
    return not len(root_g_node_alias_words) > 1


def check_is_ads1115_i2c_address(v: int) -> None:
    """
    Ads1115I2cAddress: v [0x48, 0x49, 0x4a, 0x4b].

    One of the 4 allowable I2C addresses for Texas Instrument Ads1115 chips.

    Raises:
        ValueError: if not Ads1115I2cAddress format
    """
    if v not in [0x48, 0x49, 0x4A, 0x4B]:
        raise ValueError(f"Not Ads1115I2cAddress: <{hex(v)}>")


def check_is_near5(v: str | float) -> None:
    """
    4.5  <= v  <= 5.5
    """
    v = float(v)
    min_pi_voltage = 4.5
    max_pi_voltage = 5.5
    if v < min_pi_voltage or v > max_pi_voltage:
        raise ValueError(f"<{v}> is not between 4.5 and 5.5, not Near5")


def is_bit(candidate: int) -> int:
    if candidate not in (0, 1):
        raise ValueError(f"Candidate must be 0 or 1, Got {candidate}")
    return candidate


def is_market_name(v: str) -> str:
    try:
        x = v.split(".")
    except AttributeError as e:
        raise ValueError(f"{v} failed to split on '.'") from e
    if len(x) < 3:
        raise ValueError("MarketNames need at least 3 words")
    if x[0] not in {"e", "r", "d"}:
        raise ValueError(
            f"{v} first word must be e,r or d (energy, regulation, distribution)"
        )
    if x[1] not in MarketTypeName.values():
        raise ValueError(f"{v} not recognized MarketType")
    g_node_alias = ".".join(x[2:])
    is_left_right_dot(g_node_alias)
    return v


MarketMinutes: dict[MarketTypeName, int] = {
    MarketTypeName.da60: 60,
    MarketTypeName.rt15gate5: 15,
    MarketTypeName.rt30gate5: 30,
    MarketTypeName.rt5gate5: 5,
    MarketTypeName.rt60gate30: 60,
    MarketTypeName.rt60gate30b: 60,
    MarketTypeName.rt60gate5: 60,
}


def is_market_slot_name(v: str) -> str:
    """
    MaketSlotNameLrdFormat: the format of a MarketSlotName.
      - First word must be e, r or d (energy, regulation, distribution)
      - The second word must be a MarketTypeName
      - The last word (unix time of market slot start) must
      be a 10-digit integer divisible by 300 (i.e. all MarketSlots
      start at the top of 5 minutes)
      - More strictly, the last word must be the start of a
      MarketSlot for that MarketType (i.e. divisible by 3600
      for hourly markets)
      - The middle words have LeftRightDot format (GNodeAlias
      of the MarketMaker)
    Example: e.rt60gate5.d1.isone.ver.keene.1673539200

    """
    try:
        x = v.split(".")
    except AttributeError:
        raise ValueError(f"{v} failed to split on '.'")
    slot_start_str = x[-1]
    if len(slot_start_str) != 10:
        raise ValueError(f"slot start {slot_start_str} not of length 10")
    try:
        slot_start = int(slot_start_str)
    except ValueError:
        raise ValueError(f"slot start {slot_start_str} not an int")
    is_market_name(".".join(x[:-1]))
    market_type_name = MarketTypeName(x[1])
    market_duration_minutes = MarketMinutes[market_type_name]
    if not slot_start % (market_duration_minutes * 60) == 0:
        raise ValueError(
            f"market_slot_start_s mod {market_duration_minutes * 60} must be 0"
        )
    return v


Bit = Annotated[int, BeforeValidator(is_bit)]
HandleName = Annotated[str, BeforeValidator(is_handle_name)]
HexChar = Annotated[str, BeforeValidator(is_hex_char)]
LeftRightDotStr = Annotated[str, BeforeValidator(is_left_right_dot)]
MarketName = Annotated[str, BeforeValidator(is_market_name)]
MarketSlotName = Annotated[str, BeforeValidator(is_market_slot_name)]
SpaceheatName = Annotated[str, BeforeValidator(is_spaceheat_name)]
UUID4Str = Annotated[str, BeforeValidator(is_uuid4_str)]
UTCSeconds = Annotated[
    int, Field(ge=UTC_2000_01_01_TIMESTAMP, le=UTC_3000_01_01_TIMESTAMP)
]
UTCMilliseconds = Annotated[
    int, Field(ge=UTC_2000_01_01_TIMESTAMP * 1000, le=UTC_3000_01_01_TIMESTAMP * 1000)
]



# ruff: noqa: F405, F403

from gwproto.named_types import *

from .event import *
from .misc import *

__all__ = [
    "Ack",
    "AnalogDispatch",
    "AnyEvent",
    "ChannelReadings",
    "CommEvent",
    "EventBase",
    "EventMessage",
    "EventT",
    "MQTTConnectEvent",
    "MQTTConnectFailedEvent",
    "MQTTDisconnectEvent",
    "MQTTFullySubscribedEvent",
    "MachineStates",
    "PeerActiveEvent",
    "Ping",
    "PingMessage",
    "PowerWatts",
    "ProblemEvent",
    "Problems",
    "Report",
    "ReportEvent",
    "ResponseTimeoutEvent",
    "SendSnap",
    "ShutdownEvent",
    "SingleReading",
    "StartupEvent",
    "SyncedReadings",
    "TicklistHall",
    "TicklistHallReport",
    "TicklistReed",
    "TicklistReedReport",
]



from enum import auto

from gw.enums import GwStrEnum


class ChangeStoreFlowRelay(GwStrEnum):
    """
    Events that trigger changing StoreFlowDirection finite state machine
    Values:
      - DischargeStore
      - ChargeStore

    For more information:
      - [ASLs](https://gridworks-type-registry.readthedocs.io/en/latest/)
      - [Global Authority](https://gridworks-type-registry.readthedocs.io/en/latest/enums.html#changestoreflowrelay)
    """

    DischargeStore = auto()
    ChargeStore = auto()

    @classmethod
    def default(cls) -> "ChangeStoreFlowRelay":
        return cls.DischargeStore

    @classmethod
    def values(cls) -> list[str]:
        return [elt.value for elt in cls]

    @classmethod
    def enum_name(cls) -> str:
        return "change.store.flow.relay"

    @classmethod
    def enum_version(cls) -> str:
        return "000"



# ruff: noqa: ANN401

import abc
import inspect
import re
import sys
from abc import abstractmethod
from collections.abc import Sequence

# Static analysis (mypy, pycharm) thinks 'types' here is gwproto.named_types.
from types import ModuleType  # noqa
from typing import (
    Any,
    Generic,
    Literal,
    Optional,
    TypeVar,
    Union,
    get_origin,
)

import pydantic
from pydantic import BaseModel, Field, ValidationError, create_model
from pydantic_core import ErrorDetails

from gwproto.message import Message
from gwproto.messages import AnyEvent
from gwproto.named_types import ComponentAttributeClassGt, ComponentGt
from gwproto.topic import MQTTTopic

MessageDiscriminator = TypeVar("MessageDiscriminator", bound=Message[Any])

TYPE_NAME_FIELD: str = "TypeName"


class MQTTCodec(abc.ABC):
    ENCODING = "utf-8"
    message_model: type[Message[Any]]

    def __init__(self, message_model: type[Message[Any]]) -> None:
        self.message_model = message_model

    def encode(self, content: bytes | BaseModel) -> bytes:  # noqa
        if isinstance(content, bytes):
            encoded = content
        else:
            encoded = content.model_dump_json().encode()
        return encoded

    @classmethod
    def get_unrecognized_payload_error(
        cls, e: ValidationError
    ) -> Optional[ErrorDetails]:
        for error in e.errors():
            if error.get("type", "") == "union_tag_invalid" and error.get("loc") == (
                "Payload",
            ):
                ctx = error.get("ctx", {})
                if ctx.get("discriminator") == "'TypeName'":
                    return error
        return None

    def handle_unrecognized_payload(  # noqa
        self, payload: bytes, e: ValidationError, details: ErrorDetails
    ) -> Message[Any]:
        if details.get("ctx", {}).get("tag", "").startswith("gridworks.event"):
            try:
                return Message[AnyEvent].model_validate_json(payload)
            except ValidationError as e2:
                raise e2 from e
        raise e

    def decode(self, topic: str, payload: bytes) -> Message[Any]:
        self.validate_topic(topic)
        try:
            message = self.message_model.model_validate_json(payload)
        except pydantic.ValidationError as e:
            # ValidationError can result because we receive a TypeName we don't
            # recognize, either because the sender is newer or older than our
            # code. In some cases we can still meaningfully interpret the
            # message, for example if we can recognize that it is an event
            # message, as is done here.
            if error_details := self.get_unrecognized_payload_error(e):
                return self.handle_unrecognized_payload(payload, e, error_details)
            raise
        return message

    def validate_topic(self, topic: str) -> None:
        decoded_topic = MQTTTopic.decode(topic)
        if decoded_topic.envelope_type != self.message_model.type_name():
            raise ValueError(
                f"Type {decoded_topic.envelope_type} not recognized. "
                f"Available decoders: {self.message_model.type_name()}"
            )
        self.validate_source_and_destination(decoded_topic.src, decoded_topic.dst)

    @abstractmethod
    def validate_source_and_destination(self, src: str, dst: str) -> None: ...

    @classmethod
    def _try_message_as_event(
        cls, payload: bytes, original_exception: ValidationError
    ) -> Message[Any]:
        for error in original_exception.errors():
            if error.get("type", "") == "union_tag_invalid":
                ctx = error.get("ctx", {})
                if ctx.get("discriminator", "") == "'TypeName'" and ctx.get(
                    "tag", ""
                ).startswith("gridworks.event"):
                    try:
                        return Message[AnyEvent].model_validate_json(payload)
                    except Exception as e2:  # noqa: BLE001
                        raise e2 from original_exception
        raise original_exception


def get_model_type_name(cls: Any) -> str:
    """Return cls.TypeName, if cls inherits from BaseModel, has a TypeName field,
    and that field is a Literal, else returns an empty string.
    """
    if (
        issubclass(cls, BaseModel)
        and TYPE_NAME_FIELD in cls.model_fields
        and get_origin(cls.model_fields[TYPE_NAME_FIELD].annotation) == Literal
    ):
        return str(cls.model_fields[TYPE_NAME_FIELD].default)
    return ""


def get_candidate_modules(
    module_names: str | Sequence[str],
    modules: Optional[Sequence[Any]] = None,
) -> Sequence[ModuleType]:
    if isinstance(module_names, str):
        module_names = [module_names] if module_names else []
    if unimported := [
        module_name for module_name in module_names if module_name not in sys.modules
    ]:
        raise ValueError(f"ERROR. modules {unimported} have not been imported.")
    if modules is None:
        modules = []
    return [sys.modules[module_name] for module_name in module_names] + list(modules)


EXCLUDED_TYPE_NAMES: set[str] = {Message.type_name()}


def get_candidate_payload_classes(
    module: ModuleType,
) -> Sequence[tuple[str, type[BaseModel]]]:
    """From a given module, return list of (TypeName, class) tuples for each
    object in the module that:
        * Is a class
        * Is or inherits from BaseModel
        * Has a TypeName which is a Literal with a value.
        * TypeName is not in EXCLUDED_TYPE_NAMES, which can happen if the class
          is a refinement of Message (e.g. PingMessage).
    """
    return [
        (str(module_class.model_fields[TYPE_NAME_FIELD].default), module_class)
        for _, module_class in inspect.getmembers(module, inspect.isclass)
        if (
            issubclass(module_class, BaseModel)
            and TYPE_NAME_FIELD in module_class.model_fields
            and get_origin(module_class.model_fields[TYPE_NAME_FIELD].annotation)
            == Literal
            and len(str(module_class.model_fields[TYPE_NAME_FIELD].default)) > 0
            and str(module_class.model_fields[TYPE_NAME_FIELD].default)
            not in EXCLUDED_TYPE_NAMES
        )
    ]


def include_candidate_class(
    type_name: str,
    candidate_class: type[BaseModel],
    accumulated_types: dict[str, type[BaseModel]],
    type_name_regex: Optional[re.Pattern[str]],
) -> bool:
    if (
        type_name not in EXCLUDED_TYPE_NAMES
        and type_name in accumulated_types
        and accumulated_types[type_name] is not candidate_class
    ):
        raise ValueError(
            f"ERROR {TYPE_NAME_FIELD} ({type_name}) "
            f"for {candidate_class} already seen for "
            f"class {accumulated_types[type_name]}"
        )
    return not (type_name_regex is not None and not type_name_regex.match(type_name))


def pydantic_named_types(
    module_names: str | Sequence[str],
    modules: Optional[Sequence[Any]] = None,
    type_name_regex: Optional[re.Pattern[str]] = None,
) -> list[Any]:
    """Find Pyantic BaseModels with Literal 'TypeName' fields."""
    named_types = []
    accumulated_types: dict[str, Any] = {}
    for module in get_candidate_modules(module_names, modules):
        for type_name, candidate_class in get_candidate_payload_classes(module):
            if (
                type_name in accumulated_types
                and accumulated_types[type_name] is not candidate_class
            ):
                raise ValueError(
                    f"ERROR {TYPE_NAME_FIELD} ({type_name}) "
                    f"for {candidate_class} already seen for "
                    f"class {accumulated_types[type_name]}"
                )
            if type_name_regex is None or type_name_regex.match(type_name):
                accumulated_types[type_name] = candidate_class
                named_types.append(candidate_class)
    return named_types


def create_message_model(
    model_name: str,
    module_names: str | Sequence[str] = "",
    modules: Optional[Sequence[Any]] = None,
    explicit_types: Optional[Sequence[Any]] = None,
    type_name_regex: Optional[re.Pattern[str]] = None,
) -> type[Message[Any]]:
    used_types = pydantic_named_types(
        module_names=module_names,
        modules=modules,
        type_name_regex=type_name_regex,
    )
    if explicit_types is not None:
        used_types.extend(explicit_types)
    return create_model(
        model_name,
        __base__=Message,
        Payload=(
            Union[tuple(used_types)],
            Field(..., discriminator=TYPE_NAME_FIELD),
        ),
    )


WrappedT = TypeVar("WrappedT")


class UnionWrapper(BaseModel, Generic[WrappedT]):
    """A utility class suitable for decoding a from union of types. In order
    to treat input as a union of types, Pydantic requires that our union be in
    named field, which by convention we call "Wrapped".
    """

    Wrapped: WrappedT

    @classmethod
    def create(
        cls,
        model_name: str,
        *,
        module_names: str | Sequence[str] = "",
        modules: Optional[Sequence[Any]] = None,
        explicit_types: Optional[Sequence[Any]] = None,
        type_name_regex: Optional[re.Pattern[str]] = None,
    ) -> type["UnionWrapper[WrappedT]"]:
        """Create pydantic model that is a union of all appropriate types found via
        module_names, modules and explicit_types"""
        used_types = pydantic_named_types(
            module_names=module_names,
            modules=modules,
            type_name_regex=type_name_regex,
        )
        if explicit_types is not None:
            used_types.extend(explicit_types)
        if len(used_types) == 1:
            payload_field_type = (
                used_types[0],
                Field(...),
            )
        else:
            payload_field_type = (
                Union[tuple(used_types)],
                Field(..., discriminator=TYPE_NAME_FIELD),
            )
        # Pydantic requires us to put our discriminated union in a named field.
        # We use the name 'Wrapped'.
        return create_model(
            model_name, __base__=UnionWrapper, Wrapped=payload_field_type
        )


class UnionDecoder:
    """A Utility base class for decoding from a union of types."""

    loader: type[UnionWrapper[Any]]

    def __init__(
        self,
        model_name: str,
        *,
        module_names: str | Sequence[str] = "",
        modules: Optional[Sequence[Any]] = None,
        explicit_types: Optional[Sequence[Any]] = None,
        type_name_regex: Optional[re.Pattern[str]] = None,
    ) -> None:
        self.loader = UnionWrapper.create(
            model_name=model_name,
            module_names=module_names,
            modules=modules,
            explicit_types=explicit_types,
            type_name_regex=type_name_regex,
        )


class CacDecoder(UnionDecoder):
    TYPE_NAME_REGEX = re.compile(r".*\.cac\.gt")
    loader: type[UnionWrapper[Any]]

    def __init__(
        self,
        model_name: str,
        type_name_regex: Optional[re.Pattern[str]] = TYPE_NAME_REGEX,
        **kwargs: Any,
    ) -> None:
        super().__init__(model_name, type_name_regex=type_name_regex, **kwargs)

    def decode(
        self, cac_dict: dict[str, Any], *, allow_missing: bool = True
    ) -> ComponentAttributeClassGt:
        decoded: ComponentAttributeClassGt
        try:
            decoded = self.loader.model_validate({"Wrapped": cac_dict}).Wrapped
            if not isinstance(decoded, ComponentAttributeClassGt):
                raise TypeError(
                    f"ERROR. CacDecoder decoded type {type(decoded)}, "
                    "not ComponentAttributeClassGt"
                )
        except ValidationError as e:
            if allow_missing and any(
                error.get("type") == "union_tag_invalid" for error in e.errors()
            ):
                decoded = ComponentAttributeClassGt(**cac_dict)
            else:
                raise
        return decoded


class ComponentDecoder(UnionDecoder):
    TYPE_NAME_REGEX = re.compile(r".*\.?component\.gt")

    def __init__(
        self,
        model_name: str,
        type_name_regex: Optional[re.Pattern[str]] = TYPE_NAME_REGEX,
        **kwargs: Any,
    ) -> None:
        super().__init__(model_name, type_name_regex=type_name_regex, **kwargs)

    def decode(
        self, component_dict: dict[str, Any], *, allow_missing: bool = True
    ) -> ComponentGt:
        decoded: ComponentGt
        try:
            # Pydantic requires that our union of types (components here) be in
            # a named field, which by convention we call "Wrapped".
            decoded = self.loader.model_validate({"Wrapped": component_dict}).Wrapped
            if not isinstance(decoded, ComponentGt):
                raise TypeError(
                    f"ERROR. ComponentDecoder decoded type {type(decoded)}, "
                    "not ComponentGt"
                )
        except ValidationError as e:
            if allow_missing and any(
                error.get("type") == "union_tag_invalid" for error in e.errors()
            ):
                decoded = ComponentGt(**component_dict)
            else:
                raise
        return decoded



"""Type channel.config, version 000"""

from typing import Optional

from pydantic import BaseModel, ConfigDict, PositiveInt, StrictInt, model_validator
from typing_extensions import Self

from gwproto.enums import Unit
from gwproto.property_format import (
    SpaceheatName,
)


class ChannelConfig(BaseModel):
    ChannelName: SpaceheatName
    PollPeriodMs: Optional[PositiveInt] = None
    CapturePeriodS: PositiveInt
    AsyncCapture: bool
    AsyncCaptureDelta: Optional[PositiveInt] = None
    Exponent: StrictInt
    Unit: Unit
    TypeName: str = "channel.config"
    Version: str = "000"

    model_config = ConfigDict(use_enum_values=True)

    def __hash__(self) -> int:
        return hash(self.ChannelName)

    @model_validator(mode="after")
    def check_axiom_1(self) -> Self:
        """
        Axiom 1: Capture and Polling Consistency.
        CapturePeriodMs (CapturePeriodS * 1000) must be larger than PollPeriodMs. If CapturePeriodMs < 10 * PollPeriodMs then CapturePeriodMs must be a multiple of PollPeriodMs.
        """
        # Implement check for axiom 2"
        return self



from enum import auto

from gw.enums import GwStrEnum


class AlertPriority(GwStrEnum):
    """
    The GridWorks in-house enum for prioritizing alerts in a context where distributed equipment
    maintenance affects people's comfort and well-being.
    Values:
      - P1Critical: Critical alerts represent situations where immediate action is required
        to prevent significant impact or harm. These alerts should be addressed immediately,
        as they have the potential to cause significant disruption or endangerment to people's
        comfort and well-being.
      - P2High: High-priority alerts represent situations that require urgent attention
        to prevent significant impact or harm. While not as critical as P1 alerts, P2 alerts
        still require prompt action to address the issue and prevent further escalation.
      - P3Medium: Medium-priority alerts represent situations that require attention to
        prevent potential impact or harm. These alerts should be addressed in a timely manner
        to prevent any negative impact on people's comfort and well-being.
      - P4Low: Low-priority alerts represent situations that require attention but are
        not urgent. While not as critical as P1, P2, or P3 alerts, P4 alerts should still be
        addressed in a reasonable timeframe to prevent any negative impact on people's comfort
        and well-being.
      - P5Info: Informational alerts provide general information and do not require immediate
        action. These alerts may include notifications, status updates, or routine maintenance
        reminders.

    For more information:
      - [ASLs](https://gridworks-type-registry.readthedocs.io/en/latest/)
      - [Global Authority](https://gridworks-type-registry.readthedocs.io/en/latest/enums.html#alertpriority)
    """

    P1Critical = auto()
    P2High = auto()
    P3Medium = auto()
    P4Low = auto()
    P5Info = auto()

    @classmethod
    def default(cls) -> "AlertPriority":
        return cls.P3Medium

    @classmethod
    def values(cls) -> list[str]:
        return [elt.value for elt in cls]

    @classmethod
    def enum_name(cls) -> str:
        return "alert.priority"

    @classmethod
    def enum_version(cls) -> str:
        return "000"



from enum import auto

from gw.enums import GwStrEnum


class RelayWiringConfig(GwStrEnum):
    """
    While some relays come with only two terminals and a default configuration, many come with
    a common terminal (COM), normally open terminal (NO) and normally closed terminal (NC).
    When the relay is de-energized, the circuit between COM and Normally Closed is closed. When
    the relay is energized, the circuit between COM and Normally Open is closed. This enum is
    about how one wires such a relay into a circuit.
    Values:
      - NormallyClosed: When the relay is de-energized, the circuit is closed (circuit
        is wired through COM and NC).
      - NormallyOpen: When the relay is de-energized, the circuit is open (circuit is
        wired through COM and NC).
      - DoubleThrow: COM, NC, and NO are all connected to parts of the circuit. For example,
        NC could activate a heat pump and NO could activate a backup oil boiler. The Double
        Throw configuration allows for switching between these two.

    For more information:
      - [ASLs](https://gridworks-type-registry.readthedocs.io/en/latest/)
      - [Global Authority](https://gridworks-type-registry.readthedocs.io/en/latest/enums.html#relaywiringconfig)
      - [More Info](https://gridworks-protocol.readthedocs.io/en/latest/relays.html)
    """

    NormallyClosed = auto()
    NormallyOpen = auto()
    DoubleThrow = auto()

    @classmethod
    def default(cls) -> "RelayWiringConfig":
        return cls.NormallyClosed

    @classmethod
    def values(cls) -> list[str]:
        return [elt.value for elt in cls]

    @classmethod
    def enum_name(cls) -> str:
        return "relay.wiring.config"

    @classmethod
    def enum_version(cls) -> str:
        return "000"



"""Type analog.dispatch, version 000"""

from typing import Literal, Optional

from pydantic import BaseModel, StrictInt, model_validator
from typing_extensions import Self

from gwproto.property_format import (
    HandleName,
    LeftRightDotStr,
    SpaceheatName,
    UTCMilliseconds,
    UUID4Str,
)


class AnalogDispatch(BaseModel):
    FromGNodeAlias: Optional[LeftRightDotStr] = None
    FromHandle: HandleName
    ToHandle: HandleName
    AboutName: SpaceheatName
    Value: StrictInt
    TriggerId: UUID4Str
    UnixTimeMs: UTCMilliseconds
    TypeName: Literal["analog.dispatch"] = "analog.dispatch"
    Version: str = "000"

    @model_validator(mode="after")
    def check_axiom_1(self) -> Self:
        """
        Axiom 1: FromHandle must be the immediate boss of ToHandle, unless ToHandle contains 'multiplexer'.

        """
        if "multiplexer" in self.ToHandle:
            return self
        if ".".join(self.ToHandle.split(".")[:-1]) != self.FromHandle:
            raise ValueError(
                f"FromHandle {self.FromHandle} must be direct boss of ToHandle {self.ToHandle}"
            )
        return self



class SchemaError(Exception):
    """Base class for Schema errors"""


class AlgoError(Exception):
    """Base class for errors related to Algorand"""


class DcError(Exception):
    """Base class for dataclass errors"""


class RegistryError(Exception):
    """Base class for registry errors"""



from gwproto.data_classes.components.component import Component
from gwproto.named_types.component_attribute_class_gt import ComponentAttributeClassGt
from gwproto.named_types.web_server_component_gt import WebServerComponentGt
from gwproto.named_types.web_server_gt import WebServerGt


class WebServerComponent(Component[WebServerComponentGt, ComponentAttributeClassGt]):
    @property
    def web_server_gt(self) -> WebServerGt:
        return self.gt.WebServer



# Literal Enum:
#  - no additional values can be added over time.
#  - Sent as-is, not in hex symbol
from enum import auto

from gwproto.enums.relay_event_base import RelayEventBase


class ChangeHeatcallSource(RelayEventBase):
    """
    Used for dispatch/actions between the two states of a double-throw failsafe relay that toggles
    between the SCADA and a Thermostat for controlling the heat call to a zone.
    """

    SwitchToWallThermostat = auto()
    SwitchToScada = auto()

    @classmethod
    def values(cls) -> list[str]:
        """
        Returns enum choices
        """
        return [elt.value for elt in cls]

    @classmethod
    def default(cls) -> "ChangeHeatcallSource":
        return cls.SwitchToWallThermostat

    @classmethod
    def enum_name(cls) -> str:
        return "change.heatcall.source"



from enum import auto

from gw.enums import GwStrEnum


class MakeModel(GwStrEnum):
    """
    Determines Make/Model of device associated to a Spaceheat Node supervised by SCADA
    Values:
      - UnknownMake__UnknownModel
      - Egauge__4030: A power meter in Egauge's 403x line. [More Info](https://drive.google.com/drive/u/0/folders/1abJ-o9tlTscsQpMvT6SHxIm5j5aODgfA).
      - NCD__PR8-14-SPST: NCD's 4-channel high-power relay controller + 4 GPIO with I2C
        interface. [More Info](https://store.ncd.io/product/4-channel-high-power-relay-controller-4-gpio-with-i2c-interface/?attribute_pa_choose-a-relay=20-amp-spdt).
      - Adafruit__642: Adafruit's high-temp, water-proof 1-wire temp sensor. [More Info](https://www.adafruit.com/product/642).
      - GridWorks__TSnap1: Actual GridWorks TSnap 1.0 SCADA Box.
      - GridWorks__WaterTempHighPrecision: PlaceHolder for some new GridWorks designed
        device.
      - Gridworks__SimPm1: Simulated power meter.
      - SchneiderElectric__Iem3455: Schneider Electric IEM 344 utility meter.
      - GridWorks__SimBool30AmpRelay: Simulated relay.
      - OpenEnergy__EmonPi: Open Energy's open source multipurpose sensing device (including
        internal power meter). [More Info](https://docs.openenergymonitor.org/emonpi/technical.html).
      - GridWorks__SimTSnap1: Simulated SCADA Box.
      - Atlas__EzFlo: Atlas Scientific EZO Embedded Flow Meter Totalizer, pulse to I2C. [More Info](https://drive.google.com/drive/u/0/folders/142bBV1pQIbMpyIR_0iRUr5gnzWgknOJp).
      - Hubitat__C7__LAN1: This refers to a Hubitat C7 that has been configured in a specific
        way with respect to the APIs it presents on the Local Area Network. The Hubitat C7 is
        a home automation hub that supports building ZigBee and ZWave meshes, plugs into Ethernet,
        has a reasonable user interface and has an active community of open-source developers
        who create drivers and package managers for devices, and supports the creation of various
        types of APIs on the Local Area Network. [More Info](https://drive.google.com/drive/folders/1AqAU_lC2phzuI9XRYvogiIYA7GXNtlr6).
      - GridWorks__Tank_Module_1: This refers to a small module designed and assembled
        by GridWorks that is meant to be mounted to the side of a hot water tank. It requires
        24V DC and has 4 temperature sensors coming out of it labeled 1, 2, 3 and 4. It is meant
        to provide temperature readings (taken within a half a second of each other) of all
        4 of its sensors once a minute. [More Info](https://drive.google.com/drive/folders/1GSxDd8Naf1GKK_fSOgQU933M1UcJ4r8q).
      - Fibaro__Analog_Temp_Sensor: This enum refers to a Fibaro FGBS-222 home automation
        device that has been configured in a specific way. This includes (1) being attached
        to two 10K NTC thermistors and a specific voltage divider circuit that specifies its
        temperature as a function of voltage and (2) one of its potential free outputs being
        in-line with the power of a partner Fibaro, so that it can power cycle its partner (because
        there are reports of Fibaros no longer reporting temp change after weeks or months until
        power cylced). The Fibaro itself is a tiny (29 X 18 X 13 mm) Z-Wave device powered on
        9-30V DC that can read up to 6 1-wire DS18B20 temp sensors, 2 0-10V analog inputs and
        also has 2 potential free outputs. [More Info](https://drive.google.com/drive/u/0/folders/1Muhsvw00goppHIfGSEmreX4hM6V78b-m).
      - Amphenol__NTC_10K_Thermistor_MA100GG103BN: A small gauge, low-cost, rapid response
        NTC 10K Thermistor designed for medical applications. [More Info](https://drive.google.com/drive/u/0/folders/11HW4ov66UvxKAwqApW6IrtoXatZBLQkd).
      - YHDC__SCT013-100: YHDC current transformer. [More Info](https://en.yhdc.com/product/SCT013-401.html).
      - Magnelab__SCT-0300-050: Magnelab 50A current transformer.
      - GridWorks__MultiTemp1: GridWorks ADS 1115-based analog temperature sensor that
        has 12 channels (labeled 1-12) to read 12 10K NTC Thermistors. It is comprised of 3
        NCD ADS 1115 I2C temperature sensors with I2C Addresses 0x4b, 0x48, 0x49. [More Info](https://drive.google.com/drive/u/0/folders/1OuY0tunaad2Ie4Id3zFB7FcbEwHizWuL).
      - Krida__Emr16-I2c-V3: 16-Channel I2C Low Voltage Electromagnetic Relay Board. [More Info](https://drive.google.com/drive/u/0/folders/1jL82MTRKEh9DDmxJFQ2yU2cjqnVD9Ik7).
      - Omega__FTB8007HW-PT: A double-jet reed pulse producing Flow Meter with 3/4" pipe,
        one pulse per 1/10th of a gallon. [More Info](https://drive.google.com/drive/u/0/folders/1gPR4nIGUuEVyBqBjb2wfY1Znqh6MvKWw).
      - Istec_4440: A double-jet reed pulse producing Flow Meter with 3/4" pipe, somewhat
        strange pulse output. [More Info](https://drive.google.com/drive/u/0/folders/1nioNO_XeEzE4NQJKXvuFq74_HH1vwRc6).
      - Omega__FTB8010HW-PT: A double-jet reed pulse producingFlow Meter with 1" pipe,
        one pulse per gallon. Rated for water to 195F. [More Info](https://drive.google.com/drive/u/0/folders/1fiFr9hwYGeXZ1SmpxaSz_XROhfThGbq8).
      - Belimo__BallValve232VS: Belimo Ball Valve. Configurable to be either normally
        open or normally closed. Goes into its powered position over about a minute and winds
        up a spring as it does that. Moves back to un-powered position in about 20 seconds, [More Info](https://drive.google.com/drive/u/0/folders/1eTqPNKaKzjKSWwnvY36tZkkv4WVdvrR3).
      - Belimo__DiverterB332L: Belimo 3-way diverter valve, 1.25", 24 VAC, spring return
        actuator. [More Info](https://drive.google.com/drive/u/0/folders/1YF_JdUoXrT3bDoXvEwqEvAi7EjahErHk).
      - Taco__0034ePLUS: Taco 0034ePLUS 010V controllable pump. [More Info](https://drive.google.com/drive/u/0/folders/1GUaQnrfiJeAmmfMiZT1fjPPIXxcTtTsj).
      - Taco__007e: Taco 007e basic circulator pump. [More Info](https://drive.google.com/drive/u/0/folders/12LIMxHMFXujV7mY53IItKP3J2EaM2JlV).
      - Armstrong__CompassH: Armstrong CompassH 010V controllable pump. [More Info](https://drive.google.com/drive/u/0/folders/1lpdvjVYD9qk7AHQnRSoY9Xf_o_L0tY38).
      - Honeywell__T6-ZWave-Thermostat: Honeywell TH6320ZW2003 T6 Pro Series Z-Wave Thermostat. [More Info](https://drive.google.com/drive/u/0/folders/1mqnU95tOdeeSGA6o3Ac_sJ1juDy84BIE).
      - PRMFiltration__WM075: A double-jet reed pulse producing Flow Meter with 3/4" pipe,
        one pulse per gallon. Cheaper than omegas. [More Info](https://drive.google.com/drive/u/0/folders/1LW-8GHekH9I8vUtT7_xC_9KvkwfZBvid).
      - BellGossett__Ecocirc20_18: A 0-10V controllable pump that switches out of 0-10V
        control when sent a 0 V signal.
      - Tewa__TT0P-10KC3-T105-1500: A 10K NTC thermistor used for wrapping around water
        pipes. [More Info](https://drive.google.com/drive/u/0/folders/1lZFZbpjBFgAQ_wlnKJxmEeiN-EOV9Erl).
      - EKM__HOT-SPWM-075-HD: 3/4" horizontal hot water flow pulse meter, 1 pulse per
        1/100 cubic ft (~0.0748 gallons).
      - GridWorks__SimMultiTemp: Simulated 12-channel Ads111x-based analog temp sensor
      - GridWorks__SimTotalizer: Simulated I2c-based pulse counter.
      - Krida__Double-Emr16-I2c-V3: Two 16-Channel I2C Low Voltage Electromagnetic Relay
        Board, with first at address 0x20 and second at address 0x21
      - GridWorks__SimDouble16PinI2cRelay: Simulated I2c Relay board with two boards and
        32 pins (for dev code using Krida__Doubler-Emr16-I2c-V3).
      - GridWorks__TankModule2
      - GridWorks__PicoFlowHall
      - GridWorks__PicoFlowReed
      - Saier__Sen-HZG1WA
      - DFRobot__DFR0971_Times2: Two DfRobot DFR0971 i2c 0-10V analog output actuators,
        set so the first has address 0x5e and the second has address 0x5f

    For more information:
      - [ASLs](https://gridworks-type-registry.readthedocs.io/en/latest/)
      - [Global Authority](https://gridworks-type-registry.readthedocs.io/en/latest/enums.html#spaceheatmakemodel)
      - [More Info](https://gridworks-protocol.readthedocs.io/en/latest/make-model.html)
    """

    UNKNOWNMAKE__UNKNOWNMODEL = auto()
    EGAUGE__4030 = auto()
    NCD__PR814SPST = auto()
    ADAFRUIT__642 = auto()
    GRIDWORKS__TSNAP1 = auto()
    GRIDWORKS__WATERTEMPHIGHPRECISION = auto()
    GRIDWORKS__SIMPM1 = auto()
    SCHNEIDERELECTRIC__IEM3455 = auto()
    GRIDWORKS__SIMBOOL30AMPRELAY = auto()
    OPENENERGY__EMONPI = auto()
    GRIDWORKS__SIMTSNAP1 = auto()
    ATLAS__EZFLO = auto()
    HUBITAT__C7__LAN1 = auto()
    GRIDWORKS__TANK_MODULE_1 = auto()
    FIBARO__ANALOG_TEMP_SENSOR = auto()
    AMPHENOL__NTC_10K_THERMISTOR_MA100GG103BN = auto()
    YHDC__SCT013100 = auto()
    MAGNELAB__SCT0300050 = auto()
    GRIDWORKS__MULTITEMP1 = auto()
    KRIDA__EMR16I2CV3 = auto()
    OMEGA__FTB8007HWPT = auto()
    ISTEC_4440 = auto()
    OMEGA__FTB8010HWPT = auto()
    BELIMO__BALLVALVE232VS = auto()
    BELIMO__DIVERTERB332L = auto()
    TACO__0034EPLUS = auto()
    TACO__007E = auto()
    ARMSTRONG__COMPASSH = auto()
    HONEYWELL__T6ZWAVETHERMOSTAT = auto()
    PRMFILTRATION__WM075 = auto()
    BELLGOSSETT__ECOCIRC20_18 = auto()
    TEWA__TT0P10KC3T1051500 = auto()
    EKM__HOTSPWM075HD = auto()
    GRIDWORKS__SIMMULTITEMP = auto()
    GRIDWORKS__SIMTOTALIZER = auto()
    KRIDA__DOUBLEEMR16I2CV3 = auto()
    GRIDWORKS__SIMDOUBLE16PINI2CRELAY = auto()
    GRIDWORKS__TANKMODULE2 = auto()
    GRIDWORKS__PICOFLOWHALL = auto()
    GRIDWORKS__PICOFLOWREED = auto()
    SAIER__SENHZG1WA = auto()
    DFROBOT__DFR0971_TIMES2 = auto()
    GRIDWORKS__TANKMODULE3 = auto()
    GRIDWORKS__GW101 = auto()

    @classmethod
    def default(cls) -> "MakeModel":
        return cls.UNKNOWNMAKE__UNKNOWNMODEL

    @classmethod
    def values(cls) -> list[str]:
        return [elt.value for elt in cls]

    @classmethod
    def enum_name(cls) -> str:
        return "spaceheat.make.model"

    @classmethod
    def enum_version(cls) -> str:
        return "006"



"""Type electric.meter.channel.config, version 000"""

from typing import Literal, Optional

from gwproto.named_types.channel_config import ChannelConfig
from gwproto.named_types.egauge_register_config import (
    EgaugeRegisterConfig as EgaugeConfig,
)


class ElectricMeterChannelConfig(ChannelConfig):
    EgaugeRegisterConfig: Optional[EgaugeConfig] = None
    TypeName: Literal["electric.meter.channel.config"] = "electric.meter.channel.config"
    Version: str = "000"



"""Code for actors that use a simple rest interaction, converting the response to one or more
REST commands into a message posted to main processing thread.

"""

from functools import cached_property
from typing import Any, Literal, Optional, Self

import yarl
from pydantic import BaseModel, ConfigDict, HttpUrl, model_validator

from gwproto.utils import snake_to_camel


class URLArgs(BaseModel):
    """A container for paramters that can be passed to yarl.URL.build()"""

    scheme: Literal["http", "https", ""] = "https"
    user: Optional[str] = None
    password: Optional[str] = None
    host: str = ""
    port: Optional[int] = None
    path: str = ""
    query: Optional[list[tuple[str, str | int | float]]] = None
    fragment: str = ""
    encoded: bool = True
    model_config = ConfigDict(alias_generator=snake_to_camel, populate_by_name=True)

    @classmethod
    def dict_from_url(cls, url: str | yarl.URL) -> dict[str, Any]:
        if isinstance(url, str):
            url = yarl.URL(url)
        return {
            "scheme": url.scheme,
            "user": url.user,
            "password": url.password,
            "host": url.host,
            "port": url.port,
            "path": url.path,
            "query": list(url.query.items()),
            "fragment": url.fragment,
        }

    @classmethod
    def from_url(cls, url: str | yarl.URL) -> "URLArgs":
        return URLArgs(**cls.dict_from_url(url))


class URLConfig(BaseModel):
    """Construct a URL. Three methods are provided. They are run in order of appearance,
    each updating and/or modifying the previous method.

    The methods are:
    - 'URL', an explicit string (e.g. https://www.example.org)
    - 'URLArgs', a dictionary of arguments that will be passed to yarl.URL.build()
    - 'URLPathFormat' and 'URLPathArgs', a format string and optional parameters used to
      set the 'path' portion of a yarl.URL.

    See make_url() for implementation.
    """

    url: Optional[HttpUrl] = None
    """URL as an explicit string"""

    url_args: Optional[URLArgs] = None
    """Arguments that can be passed to yarl.URL.build()"""

    url_path_format: str = ""
    """A string or format string used for the 'path' portion of the URL.
    This string will be formatted with the contents of url_path_args.
    For example, url_path_format="a/{device_id}/b" could be used used with
    url_path_args={"device_id":1} to produce a 'path' of "a/1/b".
    See make_url() for details.
    """

    url_path_args: Optional[dict[str, str | int | float]] = None
    """A dictionary of parameters used for filling in url_path_format to
    produce the URL 'path' field. The formatting operation is done as
    url_path_format.format(**URLPathArgs). See make_url() for details.
    """
    model_config = ConfigDict(alias_generator=snake_to_camel, populate_by_name=True)

    def to_url(self) -> yarl.URL:
        url = self.make_url(self)
        if url is None:
            raise ValueError("URL cannot be None")
        return url

    @classmethod
    def make_url_args(
        cls, url_config: Optional["URLConfig"]
    ) -> Optional[dict[str, Any]]:
        if url_config is None:
            return None

        # args from self.url
        if url_config.url is None:
            url_args = {}
        else:
            url_args = dict(URLArgs.from_url(yarl.URL(str(url_config.url))))

        # args from self.url_args
        if url_config.url_args is not None:
            url_args.update(url_config.url_args)

        # args from url_path_format
        if url_config.url_path_format:
            path = url_config.url_path_format
            if url_config.url_path_args:
                path = path.format(**url_config.url_path_args)
            url_args["path"] = path

        return url_args

    @classmethod
    def make_url(cls, url_config: Optional["URLConfig"]) -> Optional[yarl.URL]:
        args = URLConfig.make_url_args(url_config)
        if args:
            return yarl.URL.build(**args)
        return None


class AioHttpClientTimeout(BaseModel):
    total: Optional[float] = None
    connect: Optional[float] = None
    sock_read: Optional[float] = None
    sock_connect: Optional[float] = None
    model_config = ConfigDict(alias_generator=snake_to_camel, populate_by_name=True)


class SessionArgs(BaseModel):
    base_url: Optional[URLConfig] = None
    timeout: Optional[AioHttpClientTimeout] = None
    model_config = ConfigDict(
        extra="allow", alias_generator=snake_to_camel, populate_by_name=True
    )


class RequestArgs(BaseModel):
    url: Optional[URLConfig] = None
    method: Literal["GET", "POST", "PUT", "DELETE"] = "GET"
    params: Optional[dict[str, Any]] = None
    data: Optional[dict[str, Any] | list[Any] | tuple[Any]] = None
    headers: Optional[dict[str, Any]] = None
    timeout: Optional[AioHttpClientTimeout] = None
    ssl: Optional[bool] = None
    model_config = ConfigDict(
        extra="allow", alias_generator=snake_to_camel, populate_by_name=True
    )


class ErrorResponse(BaseModel):
    error_for_http_status: bool = True
    raise_exception: bool = False
    report: bool = True
    model_config = ConfigDict(
        extra="allow", alias_generator=snake_to_camel, populate_by_name=True
    )


class ErrorResponses(BaseModel):
    request: ErrorResponse = ErrorResponse()
    convert: ErrorResponse = ErrorResponse()
    model_config = ConfigDict(
        extra="allow", alias_generator=snake_to_camel, populate_by_name=True
    )


DEFAULT_REST_POLL_PERIOD_SECONDS = 60.0


class RESTPollerSettings(BaseModel):
    session: SessionArgs = SessionArgs()
    request: RequestArgs = RequestArgs()
    poll_period_seconds: float = DEFAULT_REST_POLL_PERIOD_SECONDS
    errors: ErrorResponses = ErrorResponses()
    model_config = ConfigDict(
        extra="allow",
        alias_generator=snake_to_camel,
        populate_by_name=True,
        ignored_types=(cached_property,),
    )

    def url_args(self) -> dict[str, Any]:
        session_args = URLConfig.make_url_args(self.session.base_url)
        request_args = URLConfig.make_url_args(self.request.url)
        if session_args is not None:
            url_args = session_args
            if request_args is not None:
                url_args.update(request_args)
        else:
            if request_args is None:
                raise ValueError(
                    "Neither session.base_url nor request.url produces a URL"
                )
            url_args = request_args
        return url_args

    @cached_property
    def url(self) -> yarl.URL:
        session_args = URLConfig.make_url_args(self.session.base_url)
        request_args = URLConfig.make_url_args(self.request.url)
        if session_args is not None:
            url_args = session_args
            if request_args is not None:
                url_args.update(request_args)
        else:
            if request_args is None:
                raise ValueError(
                    "Neither session.base_url nor request.url produces a URL"
                )
            url_args = request_args
        return yarl.URL.build(**url_args)

    def clear_property_cache(self) -> None:
        self.__dict__.pop("url", None)

    @model_validator(mode="after")
    def post_model_validator(self) -> Self:
        base_url = URLConfig.make_url(self.session.base_url)
        url = URLConfig.make_url(self.request.url)
        if base_url is None and url is None:
            raise ValueError(
                "ERROR. At least one of session.base_url and request.url must be specified"
            )
        if base_url is None and (url is None or not url.is_absolute()):
            raise ValueError(
                "ERROR. if session.base_url is None, request.url must be absolute\n"
                f"  request.url:      <{url}>\n"
            )
        if base_url is not None and not base_url.is_absolute():
            raise ValueError(
                f"ERROR. session.base_url is not absolute.\n"
                f"  session.base_url: <{base_url}>\n"
            )
        if base_url is not None and url is not None:
            if url.is_absolute():
                raise ValueError(
                    "ERROR. Both session.base_url and request.url are absolute.\n"
                    f"  session.base_url: <{base_url}>\n"
                    f"  request.url:      <{url}>\n"
                )
            if not url.path.startswith("/"):
                raise ValueError(
                    "ERROR. If session.base_url not None, request.url.path must start with '/'.\n"
                    f"  session.base_url: <{base_url}>\n"
                    f"  request.url:      <{url}>\n"
                    f"  request.url.path: <{url.path}>\n"
                )
        return self



"""Type keyparam.change.log, version 000"""

from typing import Literal

from pydantic import BaseModel, ConfigDict, field_validator

from gwproto.enums import KindOfParam
from gwproto.property_format import (
    LeftRightDotStr,
    check_is_log_style_date_with_millis,
)


class KeyparamChangeLog(BaseModel):
    """
    Key Param Change Record.

    The keyparam.change.record type is designed for straightforward logging of important parameter
    changes in the SCADA and AtomicTNode code for transactive space-heating systems. Check out
    the details in [gridworks-atn]( https://github.com/thegridelectric/gridworks-atn) and [gw-scada-spaceheat-python](https://github.com/thegridelectric/gw-scada-spaceheat-python).
    It's made for humans—developers and system maintainers—to easily create and reference records
    of significant changes. Keep it short and sweet. We suggest using a "Before" and "After"
    attribute pattern to include the changed value, focusing for example on specific components
    rather than the entire hardware layout.
    """

    AboutNodeAlias: LeftRightDotStr
    ChangeTimeUtc: str
    Author: str
    ParamName: str
    Description: str
    Kind: KindOfParam
    TypeName: Literal["keyparam.change.log"] = "keyparam.change.log"
    Version: str = "000"

    model_config = ConfigDict(extra="allow", use_enum_values=True)

    @field_validator("ChangeTimeUtc")
    @classmethod
    def _check_change_time_utc(cls, v: str) -> str:
        try:
            check_is_log_style_date_with_millis(v)
        except ValueError as e:
            raise ValueError(
                f"ChangeTimeUtc failed LogStyleDateWithMillis format validation: {e}",
            ) from e
        return v



import time
import uuid
from enum import Enum
from typing import Any, Generic, Literal, Optional, Self, TypeVar

from pydantic import BaseModel, Field, field_validator, model_validator

from gwproto.message import Message, as_enum
from gwproto.named_types import Report
from gwproto.property_format import UTCMilliseconds


class EventBase(BaseModel):
    MessageId: str = Field(default_factory=lambda: str(uuid.uuid4()))
    TimeCreatedMs: UTCMilliseconds = Field(
        default_factory=lambda: int(time.time() * 1000)
    )
    Src: str = ""
    TypeName: str


class AnyEvent(EventBase, extra="allow"):
    MessageId: str
    TimeCreatedMs: int = Field(default_factory=lambda: int(time.time() * 1000))
    Src: str
    TypeName: str


EventT = TypeVar("EventT", bound=EventBase)


class EventMessage(Message[EventT], Generic[EventT]):
    def __init__(self, AckRequired: bool = True, **kwargs: Any) -> None:  # noqa: ANN401, FBT001, FBT002, N803
        super().__init__(AckRequired=AckRequired, **kwargs)


class StartupEvent(EventBase):
    TypeName: Literal["gridworks.event.startup"] = "gridworks.event.startup"


class ShutdownEvent(EventBase):
    Reason: str
    TypeName: Literal["gridworks.event.shutdown"] = "gridworks.event.shutdown"
    Version: str = "001"


class Problems(Enum):
    error = "error"
    warning = "warning"


class ProblemEvent(EventBase):
    ProblemType: Problems
    Summary: str
    Details: str = ""
    TypeName: Literal["gridworks.event.problem"] = "gridworks.event.problem"
    Version: str = "001"

    @field_validator("ProblemType", mode="before")
    @classmethod
    def problem_type_value(cls, v: Any) -> Optional[Problems]:  # noqa: ANN401
        return as_enum(v, Problems, Problems.error)


class CommEvent(EventBase):
    PeerName: str


class MQTTCommEvent(CommEvent): ...


class MQTTConnectEvent(MQTTCommEvent):
    TypeName: Literal["gridworks.event.comm.mqtt.connect"] = (
        "gridworks.event.comm.mqtt.connect"
    )
    Version: str = "001"


class MQTTConnectFailedEvent(MQTTCommEvent):
    TypeName: Literal["gridworks.event.comm.mqtt.connect.failed"] = (
        "gridworks.event.comm.mqtt.connect.failed"
    )
    Version: str = "001"


class MQTTDisconnectEvent(MQTTCommEvent):
    TypeName: Literal["gridworks.event.comm.mqtt.disconnect"] = (
        "gridworks.event.comm.mqtt.disconnect"
    )
    Version: str = "001"


class MQTTFullySubscribedEvent(CommEvent):
    TypeName: Literal["gridworks.event.comm.mqtt.fully.subscribed"] = (
        "gridworks.event.comm.mqtt.fully.subscribed"
    )
    Version: str = "001"


class ResponseTimeoutEvent(CommEvent):
    TypeName: Literal["gridworks.event.comm.response.timeout"] = (
        "gridworks.event.comm.response.timeout"
    )
    Version: str = "001"


class PeerActiveEvent(CommEvent):
    TypeName: Literal["gridworks.event.comm.peer.active"] = (
        "gridworks.event.comm.peer.active"
    )
    Version: str = "001"


class ReportEvent(EventBase):
    Report: Report
    TypeName: Literal["report.event"] = "report.event"
    Version: str = "002"

    @model_validator(mode="after")
    def infer_base_fields(self) -> Self:
        if str(self.Report.Version) == "001":
            self.Version = "000"
        elif self.Report.Version == "002":
            self.Version = "002"
        self.MessageId = self.Report.Id
        self.TimeCreatedMs = self.Report.MessageCreatedMs
        return self



from typing import Any

from pydantic import BaseModel

DEFAULT_WEB_SERVER_NAME = "default"


class WebServerGt(BaseModel):
    Name: str = DEFAULT_WEB_SERVER_NAME
    Host: str = "localhost"
    Port: int = 8080
    Enabled: bool = True
    Kwargs: dict[str, Any] = {}



"""type ticklist.reed.report, version 000"""

from typing import Literal

from pydantic import BaseModel  # Count:true

from gwproto.named_types.ticklist_reed import TicklistReed
from gwproto.property_format import LeftRightDotStr, SpaceheatName, UTCMilliseconds


class TicklistReedReport(BaseModel):
    """
    Used by the SCADA to forward a ticklist.reed message received from a PicoFlowReed module.
    """

    TerminalAssetAlias: LeftRightDotStr
    ChannelName: SpaceheatName
    ScadaReceivedUnixMs: UTCMilliseconds
    Ticklist: TicklistReed
    TypeName: Literal["ticklist.reed.report"] = "ticklist.reed.report"
    Version: str = "000"



"""Type relay.actor.config, version 002"""

from typing import Literal

from pydantic import PositiveInt, model_validator
from typing_extensions import Self

from gwproto.enums import RelayWiringConfig
from gwproto.named_types import ChannelConfig
from gwproto.property_format import (
    SpaceheatName,
)


class RelayActorConfig(ChannelConfig):
    """
    Relay Actor Config.

    Used to associate individual relays on a multi-channel relay board to specific SpaceheatNode
    actors. Each actor managed by the Spaceheat SCADA has an associated SpaceheatNode. That
    Node will be associated to a relay board component with multiple relays. Th relay board
    will have a list of relay actor configs so that the actor can identify which relay it has
    purview over. Has DeEnergizedState and EnergizedState

    [More info](https://gridworks-protocol.readthedocs.io/en/latest/spaceheat-actor.html)
    """

    RelayIdx: PositiveInt
    ActorName: SpaceheatName
    WiringConfig: RelayWiringConfig
    EventType: str
    DeEnergizingEvent: str
    EnergizingEvent: str
    StateType: str
    DeEnergizedState: str
    EnergizedState: str
    TypeName: Literal["relay.actor.config"] = "relay.actor.config"
    Version: str = "002"

    @model_validator(mode="after")
    def check_axiom_1(self) -> Self:
        """
        Axiom 1: EventType, DeEnergizingEvent/EnergizingEvent consistency.
        If the event type is the name of a known enum, then the DeEnergizingEvent, EnergizingEvent pair are the values of that enum.
        """
        # Implement check for axiom 1"
        return self

    @model_validator(mode="after")
    def check_axiom_2(self) -> Self:
        """
        Axiom 2: StateType, EnergizedState/DeEnergizedState consistency.
        If the state type is the name of a known enum, then the DeEnergizedState, EnergizedState pair are the values of that enum.
        """
        # Implement check for axiom 2"
        return self

    @model_validator(mode="after")
    def check_axiom_3(self) -> Self:
        """
        Axiom 3: Events and States match. .
         E.g. if RelayOpen is the EnergizedState then the EnergizingEvent is OpenRelay.
        """
        # Implement check for axiom 3"
        return self



import copy
import re
from functools import cached_property
from typing import Annotated, Optional

import yarl
from pydantic import BaseModel, ConfigDict, Field, field_validator

from gwproto.enums import TelemetryName, Unit
from gwproto.named_types.hubitat_component_gt import HubitatRESTResolutionSettings
from gwproto.named_types.rest_poller_gt import (
    DEFAULT_REST_POLL_PERIOD_SECONDS,
    RequestArgs,
    RESTPollerSettings,
    URLArgs,
    URLConfig,
)
from gwproto.utils import snake_to_camel

HUBITAT_ID_REGEX = re.compile(
    r".*/apps/api/(?P<api_id>-?\d+)/devices/(?P<device_id>-?\d+).*?"
)
HUBITAT_ACCESS_TOKEN_REGEX = re.compile(
    r".*\?.*access_token=(?P<access_token>[a-fA-F0-9\-]+).*"
)


class FibaroTempSensorSettingsGt(BaseModel):
    stack_depth: Annotated[int, Field(ge=1)]
    device_id: int
    fibaro_component_id: str
    analog_input_id: Annotated[int, Field(ge=1, le=2)]
    tank_label: str = ""
    exponent: int = 1
    telemetry_name: TelemetryName = TelemetryName.WaterTempCTimes1000
    temp_unit: Unit = Unit.Celcius
    enabled: bool = True
    web_listen_enabled: bool = True
    poll_period_seconds: Optional[float] = None
    """The actual poll_seconds_used will be the first of:
    1. Any value specified in an explicit 'rest' member, if 'rest' member is not None.
    2. The value from this object, if that value is not None.
    3. The value of HubitatTanksSettingsGt.default_poll_period_seconds if it is not None.
    4. The default value.
    """
    rest: Optional[RESTPollerSettings] = None
    model_config = ConfigDict(
        extra="allow",
        use_enum_values=True,
        alias_generator=snake_to_camel,
        populate_by_name=True,
    )


DEFAULT_SENSOR_NODE_NAME_FORMAT = "{tank_name}-depth{stack_depth}"


class FibaroTempSensorSettings(FibaroTempSensorSettingsGt):
    node_name: str
    model_config = ConfigDict(ignored_types=(cached_property, TelemetryName))

    @field_validator("rest")
    @classmethod
    def _collapse_rest_url(
        cls, v: Optional[RESTPollerSettings]
    ) -> Optional[RESTPollerSettings]:
        if v is not None:
            # Collapse session.base_url and request.url into
            # request.url.
            collapsed_url = v.url
            v.session.base_url = URLConfig()
            if v.request.url is None:
                v.request.url = URLConfig()
            v.request.url.url_args = URLArgs.from_url(collapsed_url)
        return v

    @cached_property
    def url(self) -> yarl.URL:
        if self.rest is None:
            raise ValueError("Cannot produce url since self.rest is None")
        return self.rest.url

    @cached_property
    def api_id(self) -> int:
        match = HUBITAT_ID_REGEX.match(str(self.url))
        if match is None:
            raise ValueError(
                f"url <{self.url}> does not match regex {HUBITAT_ID_REGEX}"
            )
        return int(match.group("api_id"))

    @property
    def host(self) -> str:
        if self.rest is None:
            raise ValueError("Cannot produce host since self.rest is None")
        host = self.rest.url.host
        if host is None:
            raise ValueError(f"Url {self.rest.url} has host None")
        return host

    @cached_property
    def access_token(self) -> Optional[str]:
        match = HUBITAT_ACCESS_TOKEN_REGEX.match(str(self.url))
        if match:
            return match.group("access_token")
        return None

    def clear_property_cache(self) -> None:
        if self.rest is not None:
            self.rest.clear_property_cache()
        for prop in [
            "url",
            "api_id",
            "access_token",
        ]:
            self.__dict__.pop(prop, None)

    def resolve_rest(  # noqa: C901, PLR0912
        self,
        hubitat: HubitatRESTResolutionSettings,
    ) -> None:
        # Constuct url config on top of maker api url url config
        constructed_config = copy.deepcopy(hubitat.maker_api_url_config)
        constructed_config.url_path_format += "/devices/{device_id}/refresh"
        if constructed_config.url_path_args is None:
            constructed_config.url_path_args = {}
        constructed_config.url_path_args["device_id"] = self.device_id

        if self.rest is None:
            # Since no "inline" rest configuration is present, use constructed url config
            if self.poll_period_seconds is None:
                poll_period_seconds = DEFAULT_REST_POLL_PERIOD_SECONDS
            else:
                poll_period_seconds = self.poll_period_seconds
            self.rest = RESTPollerSettings(
                poll_period_seconds=poll_period_seconds,
                request=RequestArgs(url=constructed_config),
            )
        elif self.rest.request.url is None:
            self.rest.request.url = constructed_config
        else:
            # An inline config exists; take items *not* in inline config from
            # constructed config (inline config 'wins' on disagreement)
            existing_config = self.rest.request.url
            if existing_config.url_args is None:
                existing_config.url_args = constructed_config.url_args
            else:
                if (
                    not existing_config.url_args.host
                    and constructed_config.url_args is not None
                ):
                    existing_config.url_args.host = constructed_config.url_args.host
                if existing_config.url_path_args is None:
                    existing_config.url_path_args = constructed_config.url_path_args
                else:
                    existing_config.url_path_args = dict(
                        constructed_config.url_path_args,
                        **existing_config.url_path_args,
                    )
        self.rest.clear_property_cache()

        # Verify new URL produced by combining any inline REST configuration
        # with hubitat configuration is valid.
        url_str = str(self.rest.url)
        hubitat_gt = hubitat.component_gt.Hubitat
        # check host
        if hubitat_gt.Host != self.rest.url.host:
            raise ValueError(
                "ERROR host expected to be "
                f"{hubitat_gt.Host} but host in url is "
                f"{self.rest.url.host}, from url: <{url_str}>"
            )

        # check api_id
        if hubitat_gt.MakerApiId != self.api_id:
            raise ValueError(
                "ERROR api_id expected to be "
                f"{hubitat_gt.MakerApiId} but api_id in url is "
                f"{self.api_id}, from url: <{url_str}>"
            )

        # check device_id
        id_match = HUBITAT_ID_REGEX.match(url_str)
        if not id_match:
            raise ValueError(
                f"ERROR. ID regex <{HUBITAT_ID_REGEX.pattern}> failed to match "
                f" url <{url_str}>"
            )
        found_device_id = int(id_match.group("device_id"))
        if self.device_id != found_device_id:
            raise ValueError(
                "ERROR explicit device_id is "
                f"{self.device_id} but device in url is "
                f"{found_device_id}, from url: <{url_str}>"
            )

        # check token match
        if hubitat_gt.AccessToken != self.access_token:
            raise ValueError(
                "ERROR explicit access_token is "
                f"{hubitat_gt.AccessToken} but device in url is "
                f"{self.access_token}, from url: <{url_str}>"
            )

    @classmethod
    def create(
        cls,
        tank_name: str,
        settings_gt: FibaroTempSensorSettingsGt,
        hubitat: HubitatRESTResolutionSettings,
        default_poll_period_seconds: Optional[float] = None,
        node_name_format: str = DEFAULT_SENSOR_NODE_NAME_FORMAT,
    ) -> "FibaroTempSensorSettings":
        settings = FibaroTempSensorSettings(
            node_name=node_name_format.format(
                tank_name=tank_name,
                stack_depth=settings_gt.stack_depth,
            ),
            **settings_gt.model_dump(),
        )
        if settings.poll_period_seconds is None:
            settings.poll_period_seconds = default_poll_period_seconds
        settings.resolve_rest(hubitat)
        return settings


DEFAULT_TANK_MODULE_VOLTAGE = 23.7


class HubitatTankSettingsGt(BaseModel):
    hubitat_component_id: str
    sensor_supply_voltage: float = DEFAULT_TANK_MODULE_VOLTAGE
    default_poll_period_seconds: Optional[float] = None
    devices: list[FibaroTempSensorSettingsGt] = []
    web_listen_enabled: bool = True
    model_config = ConfigDict(
        extra="allow", alias_generator=snake_to_camel, populate_by_name=True
    )



from enum import auto

from gw.enums import GwStrEnum


class AdminState(GwStrEnum):
    """

    Values:
      - Awake
      - Dormant

    For more information:
      - [ASLs](https://gridworks-type-registry.readthedocs.io/en/latest/)
      - [Global Authority](https://gridworks-type-registry.readthedocs.io/en/latest/enums.html#adminstate)
    """

    Awake = auto()
    Dormant = auto()

    @classmethod
    def default(cls) -> "AdminState":
        return cls.Dormant

    @classmethod
    def values(cls) -> list[str]:
        return [elt.value for elt in cls]

    @classmethod
    def enum_name(cls) -> str:
        return "admin.state"

    @classmethod
    def enum_version(cls) -> str:
        return "000"



"""Type ads111x.based.cac.gt, version 000"""

from typing import Literal

from pydantic import (
    ConfigDict,
    PositiveInt,
    StrictInt,
    field_validator,
    model_validator,
)
from typing_extensions import Self

from gwproto.enums import TelemetryName
from gwproto.named_types.component_attribute_class_gt import ComponentAttributeClassGt
from gwproto.property_format import (
    check_is_ads1115_i2c_address,
)


class Ads111xBasedCacGt(ComponentAttributeClassGt):
    AdsI2cAddressList: list[StrictInt]
    TotalTerminalBlocks: PositiveInt
    TelemetryNameList: list[TelemetryName]
    TypeName: Literal["ads111x.based.cac.gt"] = "ads111x.based.cac.gt"
    Version: str = "000"

    model_config = ConfigDict(extra="allow", use_enum_values=True)

    @field_validator("AdsI2cAddressList")
    @classmethod
    def _check_ads_i2c_address_list(cls, v: list[int]) -> list[int]:
        try:
            for elt in v:
                check_is_ads1115_i2c_address(elt)
        except ValueError as e:
            raise ValueError(
                f"AdsI2cAddressList element failed Ads1115I2cAddress format validation: {e}",
            ) from e
        return v

    @model_validator(mode="after")
    def check_axiom_1(self) -> Self:
        """
        Axiom 1: TerminalBlock Ads Chip consistency.
        TotalTerminalBlocks should be greater than 4 * (len(AdsI2cAddressList) - 1 ) and less than or equal to 4*len(AdsI2cAddressList)
        """
        # Implement check for axiom 1"
        return self



from typing import Literal

from gwproto.named_types import ComponentGt


class FibaroSmartImplantComponentGt(ComponentGt):
    ZWaveDSK: str = ""
    TypeName: Literal["fibaro.smart.implant.component.gt"] = (
        "fibaro.smart.implant.component.gt"
    )



"""Type power.watts, version 000"""

from typing import Literal

from pydantic import BaseModel


class PowerWatts(BaseModel):
    """
    Real-time power of TerminalAsset in Watts.

    Used by a SCADA -> Atn or Atn -> AggregatedTNode to report real-time power of their TerminalAsset.
    Positive number means WITHDRAWAL from the grid - so generating electricity creates a negative
    number. This message is considered worse than useless to send after the first attempt, and
    does not require an ack. Shares the same purpose as gs.pwr, but is not designed to minimize
    bytes so comes in JSON format.
    """

    Watts: int
    TypeName: Literal["power.watts"] = "power.watts"
    Version: str = "000"



from enum import auto

from gw.enums import GwStrEnum


class Strategy(GwStrEnum):
    """

    Values:
      - Ha2Oil: HomeAlone2 - Also a home alone strategy for House0. If the heat pump is
        not coming on, then switch to oil boiler. TODO: DESIGN, WRITE UP
      - Ha1: HomeAlone1 - the first home alone strategy for House0. TODO: WRITE UP

    For more information:
      - [ASLs](https://gridworks-type-registry.readthedocs.io/en/latest/)
      - [Global Authority](https://gridworks-type-registry.readthedocs.io/en/latest/enums.html#spaceheatstrategy)
    """

    Ha2Oil = auto()
    Ha1 = auto()

    @classmethod
    def default(cls) -> "Strategy":
        return cls.Ha1

    @classmethod
    def values(cls) -> list[str]:
        return [elt.value for elt in cls]

    @classmethod
    def enum_name(cls) -> str:
        return "spaceheat.strategy"

    @classmethod
    def enum_version(cls) -> str:
        return "000"



from pydantic import BaseModel, ConfigDict

from gwproto.enums import TelemetryName, Unit
from gwproto.utils import snake_to_camel


class MakerAPIAttributeGt(BaseModel):
    attribute_name: str
    channel_name: str
    node_name: str
    telemetry_name: TelemetryName = TelemetryName.WaterTempCTimes1000
    unit: Unit = Unit.Celcius
    exponent: int = 3
    interpret_as_number: bool = True
    enabled: bool = True
    web_poll_enabled: bool = True
    web_listen_enabled: bool = True
    report_missing: bool = True
    report_parse_error: bool = True

    model_config = ConfigDict(
        extra="allow",
        use_enum_values=True,
        alias_generator=snake_to_camel,
        populate_by_name=True,
    )


class HubitatPollerGt(BaseModel):
    hubitat_component_id: str
    device_id: int
    attributes: list[MakerAPIAttributeGt] = []
    enabled: bool = True
    web_listen_enabled: bool = True
    poll_period_seconds: float = 60
    model_config = ConfigDict(
        extra="allow", alias_generator=snake_to_camel, populate_by_name=True
    )



"""Type ticklist.reed, version 101"""

from typing import Literal, Optional

from pydantic import BaseModel, StrictInt, model_validator
from typing_extensions import Self


class TicklistReed(BaseModel):
    HwUid: str
    FirstTickTimestampNanoSecond: Optional[StrictInt] = None
    RelativeMillisecondList: list[StrictInt]
    PicoBeforePostTimestampNanoSecond: StrictInt
    TypeName: Literal["ticklist.reed"] = "ticklist.reed"
    Version: str = "101"

    @model_validator(mode="after")
    def check_axiom_1(self) -> Self:
        """
        Axiom 1: .
        FirstTickTimestampNanoSecond is None iff RelativeMillisecondList has length 0
        """
        if (
            self.FirstTickTimestampNanoSecond is None
            and len(self.RelativeMillisecondList) > 0
        ):
            raise ValueError(
                "FirstTickTimestampNanoSecond is None but RelativeMillisecondList has nonzero length!"
            )
        if self.FirstTickTimestampNanoSecond and len(self.RelativeMillisecondList) == 0:
            raise ValueError(
                "FirstTickTimestampNanoSecond exists but RelativeMillisecondList has no elements!"
            )
        return self



"""Code for actors that use a simple rest interaction, converting the response to one or more
REST commands into a message posted to main processing thread.

"""

from typing import Literal

from gwproto.named_types import ComponentGt
from gwproto.named_types.rest_poller_gt import RESTPollerSettings


class RESTPollerComponentGt(ComponentGt):
    Rest: RESTPollerSettings
    TypeName: Literal["rest.poller.component.gt"] = "rest.poller.component.gt"



"""Type ads111x.based.component.gt, version 000"""

from collections.abc import Sequence
from typing import Literal

from pydantic import ConfigDict, field_validator

from gwproto.named_types.ads_channel_config import (
    AdsChannelConfig,
)
from gwproto.named_types.component_gt import ComponentGt
from gwproto.property_format import (
    check_is_near5,
)


class Ads111xBasedComponentGt(ComponentGt):
    OpenVoltageByAds: list[float]
    ConfigList: Sequence[AdsChannelConfig]
    TypeName: Literal["ads111x.based.component.gt"] = "ads111x.based.component.gt"
    Version: str = "000"

    model_config = ConfigDict(use_enum_values=True)

    @field_validator("OpenVoltageByAds")
    @classmethod
    def _check_open_voltage_by_ads(cls, v: list[float]) -> list[float]:
        try:
            for elt in v:
                check_is_near5(elt)
        except ValueError as e:
            raise ValueError(
                f"OpenVoltageByAds element failed Near5 format validation: {e}",
            ) from e
        return v

    @field_validator("ConfigList")
    @classmethod
    def check_ads_channel_config_list(
        cls, v: Sequence[AdsChannelConfig]
    ) -> Sequence[AdsChannelConfig]:
        """
            Axiom 1: Terminal Block consistency and Channel Name uniqueness.
            Terminal Block consistency and Channel Name uniqueness. - Each TerminalBlockIdx occurs at
        most once in the ConfigList .Each data channel occurs at most once in the ConfigList
        """
        # Implement Axiom(s)
        return v



"""ResistiveHeaterComponent definition"""

from gwproto.data_classes.components.component import Component
from gwproto.named_types import ResistiveHeaterCacGt, ResistiveHeaterComponentGt


class ResistiveHeaterComponent(
    Component[ResistiveHeaterComponentGt, ResistiveHeaterCacGt]
): ...



from gwproto.data_classes.hardware_layout import HardwareLayout
from gwproto.data_classes.sh_node import ShNode
from gwproto.decoders import (
    CacDecoder,
    ComponentDecoder,
    MessageDiscriminator,
    MQTTCodec,
    create_message_model,
    pydantic_named_types,
)
from gwproto.default_decoders import (
    default_cac_decoder,
    default_component_decoder,
)
from gwproto.errors import SchemaError
from gwproto.message import Header, Message, as_enum
from gwproto.topic import DecodedMQTTTopic, MQTTTopic

__all__ = [
    "CacDecoder",
    "ComponentDecoder",
    "DecodedMQTTTopic",
    "HardwareLayout",
    "Header",
    "MQTTCodec",
    "MQTTTopic",
    "Message",
    "MessageDiscriminator",
    "SchemaError",
    "ShNode",
    "as_enum",
    "create_message_model",
    "default_cac_decoder",
    "default_component_decoder",
    "messages",
    "property_format",
    "pydantic_named_types",
]



"""Type synced.readings, version 000"""

from typing import Literal

from pydantic import BaseModel, ConfigDict, StrictInt, model_validator
from typing_extensions import Self

from gwproto.property_format import (
    SpaceheatName,
    UTCMilliseconds,
)


class SyncedReadings(BaseModel):
    ChannelNameList: list[SpaceheatName]
    ValueList: list[StrictInt]
    ScadaReadTimeUnixMs: UTCMilliseconds
    TypeName: Literal["synced.readings"] = "synced.readings"
    Version: str = "000"

    model_config = ConfigDict(use_enum_values=True)

    @model_validator(mode="after")
    def check_axiom_1(self) -> Self:
        """
        Axiom 1: list Length Consistency.
        len(ChannelNameList) = len(ValueList)
        """
        # Implement check for axiom 1"
        return self



"""Type spaceheat.node.gt, version 110"""

from typing import Literal, Optional

from pydantic import BaseModel, ConfigDict, StrictInt, model_validator
from typing_extensions import Self

from gwproto.property_format import HandleName, SpaceheatName, UUID4Str


class SpaceheatNodeGt(BaseModel):
    Name: SpaceheatName
    ActorHierarchyName: Optional[HandleName] = None
    Handle: Optional[HandleName] = None
    ActorClass: str
    DisplayName: Optional[str] = None
    ComponentId: Optional[str] = None
    NameplatePowerW: Optional[StrictInt] = None
    InPowerMetering: Optional[bool] = None
    ShNodeId: UUID4Str
    TypeName: Literal["spaceheat.node.gt"] = "spaceheat.node.gt"
    Version: str = "200"

    @model_validator(mode="after")
    def check_axiom_1(self) -> Self:
        """
        Axiom 1: InPowerMetering requirements.
        If InPowerMetering exists and is true, then NameplatePowerW must exist
        """
        if self.InPowerMetering and self.NameplatePowerW is None:
            raise ValueError(
                "Axiom 1 failed! "
                "If InPowerMetering exists and is true, then NameplatePowerW must exist"
            )
        return self

    model_config = ConfigDict(extra="allow", use_enum_values=True)



from enum import auto

from gw.enums import GwStrEnum


class FsmName(GwStrEnum):
    """
    The name of a specific Spaceheat finite state machine. That name is used as a literal enum
    for a set of staets.
    Values:
      - Unknown
      - StoreFlowDirection
      - RelayState: Finite State Machine for a normally open or normally closed relay
        whose states (Closed, Open) are enumerated by RelayClosedOrOpen.
      - RelayPinState: Finite State Machine for a relay pin with states enumerated by
        RelayEnergizationState (Energized and DeEnergized).

    For more information:
      - [ASLs](https://gridworks-type-registry.readthedocs.io/en/latest/)
      - [Global Authority](https://gridworks-type-registry.readthedocs.io/en/latest/enums.html#shfsmname)
      - [More Info](https://gridworks-protocol.readthedocs.io/en/latest/house-0.html)
    """

    Unknown = auto()
    StoreFlowDirection = auto()
    RelayState = auto()
    RelayPinState = auto()

    @classmethod
    def default(cls) -> "FsmName":
        return cls.StoreFlowDirection

    @classmethod
    def values(cls) -> list[str]:
        return [elt.value for elt in cls]

    @classmethod
    def enum_name(cls) -> str:
        return "sh.fsm.name"

    @classmethod
    def enum_version(cls) -> str:
        return "000"



"""Old School enum (uses integers)"""

from enum import Enum


class RelayEnergizationState(Enum):
    """
    The basic two-state enum for a double-throw relay where nothing is known about its energized/de-energized
    legs. This can only ever have two states. This is an old-school enum, where DeEnergized
    encodes 0 and Energized encodes 1.
    """

    DeEnergized = 0
    Energized = 1

    @classmethod
    def values(cls) -> list[int]:
        """
        Returns enum choices
        """
        return [elt.value for elt in cls]

    @classmethod
    def default(cls) -> "RelayEnergizationState":
        return cls.DeEnergized

    @classmethod
    def enum_name(cls) -> str:
        return "relay.energization.state"



"""Type pico.btu.meter.component.gt, version 000"""

from typing import Literal, Optional

from pydantic import ConfigDict, StrictInt, model_validator
from typing_extensions import Self

from gwproto.enums import (
    GpmFromHzMethod,
    HzCalcMethod,
    MakeModel,
    TempCalcMethod,
)
from gwproto.named_types.component_gt import ComponentGt
from gwproto.property_format import (
    SpaceheatName,
)


class PicoBtuMeterComponentGt(ComponentGt):
    Enabled: bool
    SerialNumber: str
    FlowChannelName: SpaceheatName
    HotChannelName: SpaceheatName
    ColdChannelName: SpaceheatName
    ReadCtVoltage: bool
    SendHz: bool
    CtChannelName: Optional[SpaceheatName] = None
    FlowMeterType: MakeModel
    HzCalcMethod: HzCalcMethod
    TempCalcMethod: TempCalcMethod
    ThermistorBeta: StrictInt
    GpmFromHzMethod: GpmFromHzMethod
    GallonsPerPulse: float
    AsyncCaptureDeltaGpmX100: StrictInt
    AsyncCaptureDeltaCelsiusX100: StrictInt
    AsyncCaptureDeltaCtVoltsX100: Optional[StrictInt] = None
    TypeName: Literal["pico.btu.meter.component.gt"] = "pico.btu.meter.component.gt"
    Version: Literal["000"] = "000"

    model_config = ConfigDict(use_enum_values=True)

    @model_validator(mode="after")
    def check_axiom_1(self) -> Self:
        """
        Axiom 1: ReadCtVoltage is True iff AsyncCaptureDeltaCtVoltsX100 exists
        """
        if self.ReadCtVoltage and not self.AsyncCaptureDeltaCtVoltsX100:
            raise ValueError(
                f"Axiom 1 violated! ReadCtVoltage {self.ReadCtVoltage} requires AsyncCaptureDeltaCtVoltsX100!"
            )
        if not self.ReadCtVoltage and self.AsyncCaptureDeltaCtVoltsX100:
            raise ValueError(
                f"Axiom 1 violated: ReadCtVoltage {self.ReadCtVoltage} means NOAsyncCaptureDeltaCtVoltsX100"
            )
        return self
